import contextlib
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    with contextlib.suppress(ImportError):
        from celery import Celery  # type: ignore


def get_celery_worker_status(app: "Celery"):
    i = app.control.inspect()
    availability = app.control.ping()
    stats = i.stats()
    registered_tasks = i.registered()
    active_tasks = i.active()
    scheduled_tasks = i.scheduled()
    return {
        "availability": availability,
        "stats": stats,
        "registered_tasks": registered_tasks,
        "active_tasks": active_tasks,
        "scheduled_tasks": scheduled_tasks,
    }



from langflow import CustomComponent
from langchain_core.language_models.base import BaseLanguageModel
from langchain.agents.agent_toolkits.vectorstore.toolkit import VectorStoreRouterToolkit
from langchain.agents import create_vectorstore_router_agent
from typing import Callable


class VectorStoreRouterAgentComponent(CustomComponent):
    display_name = "VectorStoreRouterAgent"
    description = "Construct an agent from a Vector Store Router."

    def build_config(self):
        return {
            "llm": {"display_name": "LLM"},
            "vectorstoreroutertoolkit": {"display_name": "Vector Store Router Toolkit"},
        }

    def build(self, llm: BaseLanguageModel, vectorstoreroutertoolkit: VectorStoreRouterToolkit) -> Callable:
        return create_vectorstore_router_agent(llm=llm, toolkit=vectorstoreroutertoolkit)



from typing import Callable, Optional, Union

from langchain_anthropic import ChatAnthropic
from pydantic.v1.types import SecretStr

from langflow import CustomComponent
from langflow.field_typing import BaseLanguageModel


class ChatAnthropicComponent(CustomComponent):
    display_name = "ChatAnthropic"
    description = "`Anthropic` chat large language models."
    documentation = "https://python.langchain.com/docs/modules/model_io/models/chat/integrations/anthropic"

    def build_config(self):
        return {
            "anthropic_api_key": {
                "display_name": "Anthropic API Key",
                "field_type": "str",
                "password": True,
            },
            "model_kwargs": {
                "display_name": "Model Kwargs",
                "field_type": "dict",
                "advanced": True,
            },
            "model_name": {
                "display_name": "Model Name",
                "field_type": "str",
                "advanced": False,
                "required": False,
                "options": ["claude-3-opus-20240229", "claude-3-sonnet-20240229", "claude-3-haiku-20240307"],
            },
            "temperature": {
                "display_name": "Temperature",
                "field_type": "float",
            },
            "max_tokens": {
                "display_name": "Max Tokens",
                "field_type": "int",
                "advanced": False,
                "required": False,
            },
            "top_k": {"display_name": "Top K", "field_type": "int", "advanced": True},
            "top_p": {"display_name": "Top P", "field_type": "float", "advanced": True},
        }

    def build(
        self,
        anthropic_api_key: str,
        model_kwargs: dict = {},
        model_name: str = "claude-3-opus-20240229",
        temperature: Optional[float] = None,
        max_tokens: Optional[int] = 1024,
        top_k: Optional[int] = None,
        top_p: Optional[float] = None,
    ) -> Union[BaseLanguageModel, Callable]:
        return ChatAnthropic(
            anthropic_api_key=SecretStr(anthropic_api_key),
            model_kwargs=model_kwargs,
            model_name=model_name,
            temperature=temperature,
            max_tokens=max_tokens,  # type: ignore
            top_k=top_k,
            top_p=top_p,
        )



from datetime import datetime
from typing import TYPE_CHECKING, Optional
from uuid import UUID, uuid4

from pydantic import validator
from sqlmodel import Field, Relationship, SQLModel

if TYPE_CHECKING:
    from langflow.services.database.models.user import User


class ApiKeyBase(SQLModel):
    name: Optional[str] = Field(index=True, nullable=True, default=None)
    created_at: datetime = Field(default_factory=datetime.utcnow)
    last_used_at: Optional[datetime] = Field(default=None, nullable=True)
    total_uses: int = Field(default=0)
    is_active: bool = Field(default=True)


class ApiKey(ApiKeyBase, table=True):
    id: UUID = Field(default_factory=uuid4, primary_key=True, unique=True)

    api_key: str = Field(index=True, unique=True)
    # User relationship
    # Delete API keys when user is deleted
    user_id: UUID = Field(index=True, foreign_key="user.id")
    user: "User" = Relationship(
        back_populates="api_keys",
    )


class ApiKeyCreate(ApiKeyBase):
    api_key: Optional[str] = None
    user_id: Optional[UUID] = None


class UnmaskedApiKeyRead(ApiKeyBase):
    id: UUID
    api_key: str = Field()
    user_id: UUID = Field()


class ApiKeyRead(ApiKeyBase):
    id: UUID
    api_key: str = Field()
    user_id: UUID = Field()

    @validator("api_key", always=True)
    def mask_api_key(cls, v):
        # This validator will always run, and will mask the API key
        return f"{v[:8]}{'*' * (len(v) - 8)}"



from typing import Any, Optional

from langchain.agents import AgentExecutor, ZeroShotAgent
from langchain.agents.agent_toolkits import (
    VectorStoreInfo,
    VectorStoreRouterToolkit,
    VectorStoreToolkit,
)
from langchain.agents.agent_toolkits.vectorstore.prompt import PREFIX as VECTORSTORE_PREFIX
from langchain.agents.agent_toolkits.vectorstore.prompt import ROUTER_PREFIX as VECTORSTORE_ROUTER_PREFIX
from langchain.agents.mrkl.prompt import FORMAT_INSTRUCTIONS
from langchain.base_language import BaseLanguageModel
from langchain.chains.llm import LLMChain
from langchain.sql_database import SQLDatabase
from langchain.tools.sql_database.prompt import QUERY_CHECKER
from langchain_community.agent_toolkits import SQLDatabaseToolkit
from langchain_community.agent_toolkits.json.prompt import JSON_PREFIX, JSON_SUFFIX
from langchain_community.agent_toolkits.json.toolkit import JsonToolkit
from langchain_community.agent_toolkits.sql.prompt import SQL_PREFIX, SQL_SUFFIX
from langchain_experimental.agents.agent_toolkits.pandas.prompt import PREFIX as PANDAS_PREFIX
from langchain_experimental.agents.agent_toolkits.pandas.prompt import SUFFIX_WITH_DF as PANDAS_SUFFIX
from langchain_experimental.tools.python.tool import PythonAstREPLTool

from langflow.interface.base import CustomAgentExecutor


class JsonAgent(CustomAgentExecutor):
    """Json agent"""

    @staticmethod
    def function_name():
        return "JsonAgent"

    @classmethod
    def initialize(cls, *args, **kwargs):
        return cls.from_toolkit_and_llm(*args, **kwargs)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def from_toolkit_and_llm(cls, toolkit: JsonToolkit, llm: BaseLanguageModel):
        tools = toolkit if isinstance(toolkit, list) else toolkit.get_tools()
        tool_names = list({tool.name for tool in tools})
        prompt = ZeroShotAgent.create_prompt(
            tools,
            prefix=JSON_PREFIX,
            suffix=JSON_SUFFIX,
            format_instructions=FORMAT_INSTRUCTIONS,
            input_variables=None,
        )
        llm_chain = LLMChain(
            llm=llm,
            prompt=prompt,
        )
        agent = ZeroShotAgent(
            llm_chain=llm_chain,
            allowed_tools=tool_names,  # type: ignore
        )
        return cls.from_agent_and_tools(agent=agent, tools=tools, verbose=True)

    def run(self, *args, **kwargs):
        return super().run(*args, **kwargs)


class CSVAgent(CustomAgentExecutor):
    """CSV agent"""

    @staticmethod
    def function_name():
        return "CSVAgent"

    @classmethod
    def initialize(cls, *args, **kwargs):
        return cls.from_toolkit_and_llm(*args, **kwargs)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def from_toolkit_and_llm(
        cls, path: str, llm: BaseLanguageModel, pandas_kwargs: Optional[dict] = None, **kwargs: Any
    ):
        import pandas as pd  # type: ignore

        _kwargs = pandas_kwargs or {}
        df = pd.read_csv(path, **_kwargs)

        tools = [PythonAstREPLTool(locals={"df": df})]  # type: ignore
        prompt = ZeroShotAgent.create_prompt(
            tools,
            prefix=PANDAS_PREFIX,
            suffix=PANDAS_SUFFIX,
            input_variables=["df_head", "input", "agent_scratchpad"],
        )
        partial_prompt = prompt.partial(df_head=str(df.head()))
        llm_chain = LLMChain(
            llm=llm,
            prompt=partial_prompt,
        )
        tool_names = list({tool.name for tool in tools})
        agent = ZeroShotAgent(
            llm_chain=llm_chain,
            allowed_tools=tool_names,
            **kwargs,  # type: ignore
        )

        return cls.from_agent_and_tools(agent=agent, tools=tools, verbose=True)

    def run(self, *args, **kwargs):
        return super().run(*args, **kwargs)


class VectorStoreAgent(CustomAgentExecutor):
    """Vector store agent"""

    @staticmethod
    def function_name():
        return "VectorStoreAgent"

    @classmethod
    def initialize(cls, *args, **kwargs):
        return cls.from_toolkit_and_llm(*args, **kwargs)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def from_toolkit_and_llm(cls, llm: BaseLanguageModel, vectorstoreinfo: VectorStoreInfo, **kwargs: Any):
        """Construct a vectorstore agent from an LLM and tools."""

        toolkit = VectorStoreToolkit(vectorstore_info=vectorstoreinfo, llm=llm)

        tools = toolkit.get_tools()
        prompt = ZeroShotAgent.create_prompt(tools, prefix=VECTORSTORE_PREFIX)
        llm_chain = LLMChain(
            llm=llm,
            prompt=prompt,
        )
        tool_names = list({tool.name for tool in tools})
        agent = ZeroShotAgent(
            llm_chain=llm_chain,
            allowed_tools=tool_names,
            **kwargs,  # type: ignore
        )
        return AgentExecutor.from_agent_and_tools(agent=agent, tools=tools, verbose=True, handle_parsing_errors=True)

    def run(self, *args, **kwargs):
        return super().run(*args, **kwargs)


class SQLAgent(CustomAgentExecutor):
    """SQL agent"""

    @staticmethod
    def function_name():
        return "SQLAgent"

    @classmethod
    def initialize(cls, *args, **kwargs):
        return cls.from_toolkit_and_llm(*args, **kwargs)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def from_toolkit_and_llm(cls, llm: BaseLanguageModel, database_uri: str, **kwargs: Any):
        """Construct an SQL agent from an LLM and tools."""
        db = SQLDatabase.from_uri(database_uri)
        toolkit = SQLDatabaseToolkit(db=db, llm=llm)

        # The right code should be this, but there is a problem with tools = toolkit.get_tools()
        # related to `OPENAI_API_KEY`
        # return create_sql_agent(llm=llm, toolkit=toolkit, verbose=True)
        from langchain.prompts import PromptTemplate
        from langchain.tools.sql_database.tool import (
            InfoSQLDatabaseTool,
            ListSQLDatabaseTool,
            QuerySQLCheckerTool,
            QuerySQLDataBaseTool,
        )

        llmchain = LLMChain(
            llm=llm,
            prompt=PromptTemplate(template=QUERY_CHECKER, input_variables=["query", "dialect"]),
        )

        tools = [
            QuerySQLDataBaseTool(db=db),  # type: ignore
            InfoSQLDatabaseTool(db=db),  # type: ignore
            ListSQLDatabaseTool(db=db),  # type: ignore
            QuerySQLCheckerTool(db=db, llm_chain=llmchain, llm=llm),  # type: ignore
        ]

        prefix = SQL_PREFIX.format(dialect=toolkit.dialect, top_k=10)
        prompt = ZeroShotAgent.create_prompt(
            tools=tools,  # type: ignore
            prefix=prefix,
            suffix=SQL_SUFFIX,
            format_instructions=FORMAT_INSTRUCTIONS,
        )
        llm_chain = LLMChain(
            llm=llm,
            prompt=prompt,
        )
        tool_names = list({tool.name for tool in tools})  # type: ignore
        agent = ZeroShotAgent(
            llm_chain=llm_chain,
            allowed_tools=tool_names,
            **kwargs,  # type: ignore
        )
        return AgentExecutor.from_agent_and_tools(
            agent=agent,
            tools=tools,  # type: ignore
            verbose=True,
            max_iterations=15,
            early_stopping_method="force",
            handle_parsing_errors=True,
        )

    def run(self, *args, **kwargs):
        return super().run(*args, **kwargs)


class VectorStoreRouterAgent(CustomAgentExecutor):
    """Vector Store Router Agent"""

    @staticmethod
    def function_name():
        return "VectorStoreRouterAgent"

    @classmethod
    def initialize(cls, *args, **kwargs):
        return cls.from_toolkit_and_llm(*args, **kwargs)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def from_toolkit_and_llm(
        cls, llm: BaseLanguageModel, vectorstoreroutertoolkit: VectorStoreRouterToolkit, **kwargs: Any
    ):
        """Construct a vector store router agent from an LLM and tools."""

        tools = (
            vectorstoreroutertoolkit
            if isinstance(vectorstoreroutertoolkit, list)
            else vectorstoreroutertoolkit.get_tools()
        )
        prompt = ZeroShotAgent.create_prompt(tools, prefix=VECTORSTORE_ROUTER_PREFIX)
        llm_chain = LLMChain(
            llm=llm,
            prompt=prompt,
        )
        tool_names = list({tool.name for tool in tools})
        agent = ZeroShotAgent(
            llm_chain=llm_chain,
            allowed_tools=tool_names,
            **kwargs,  # type: ignore
        )
        return AgentExecutor.from_agent_and_tools(agent=agent, tools=tools, verbose=True, handle_parsing_errors=True)

    def run(self, *args, **kwargs):
        return super().run(*args, **kwargs)


CUSTOM_AGENTS = {
    "JsonAgent": JsonAgent,
    "CSVAgent": CSVAgent,
    "VectorStoreAgent": VectorStoreAgent,
    "VectorStoreRouterAgent": VectorStoreRouterAgent,
    "SQLAgent": SQLAgent,
}



from typing import TYPE_CHECKING
from langflow.services.database.service import DatabaseService
from langflow.services.factory import ServiceFactory

if TYPE_CHECKING:
    from langflow.services.settings.service import SettingsService


class DatabaseServiceFactory(ServiceFactory):
    def __init__(self):
        super().__init__(DatabaseService)

    def create(self, settings_service: "SettingsService"):
        # Here you would have logic to create and configure a DatabaseService
        if not settings_service.settings.DATABASE_URL:
            raise ValueError("No database URL provided")
        return DatabaseService(settings_service.settings.DATABASE_URL)



import os
from typing import Any, Callable, Dict, Optional, Union

from langchain_community.chat_models.litellm import ChatLiteLLM, ChatLiteLLMException
from langflow import CustomComponent
from langflow.field_typing import BaseLanguageModel


class ChatLiteLLMComponent(CustomComponent):
    display_name = "ChatLiteLLM"
    description = "`LiteLLM` collection of large language models."
    documentation = "https://python.langchain.com/docs/integrations/chat/litellm"

    def build_config(self):
        return {
            "model": {
                "display_name": "Model name",
                "field_type": "str",
                "advanced": False,
                "required": True,
                "info": "The name of the model to use. For example, `gpt-3.5-turbo`.",
            },
            "api_key": {
                "display_name": "API key",
                "field_type": "str",
                "advanced": False,
                "required": False,
                "password": True,
            },
            "streaming": {
                "display_name": "Streaming",
                "field_type": "bool",
                "advanced": True,
                "required": False,
                "default": True,
            },
            "temperature": {
                "display_name": "Temperature",
                "field_type": "float",
                "advanced": False,
                "required": False,
                "default": 0.7,
            },
            "model_kwargs": {
                "display_name": "Model kwargs",
                "field_type": "dict",
                "advanced": True,
                "required": False,
                "default": {},
            },
            "top_p": {
                "display_name": "Top p",
                "field_type": "float",
                "advanced": True,
                "required": False,
            },
            "top_k": {
                "display_name": "Top k",
                "field_type": "int",
                "advanced": True,
                "required": False,
            },
            "n": {
                "display_name": "N",
                "field_type": "int",
                "advanced": True,
                "required": False,
                "info": "Number of chat completions to generate for each prompt. "
                "Note that the API may not return the full n completions if duplicates are generated.",
                "default": 1,
            },
            "max_tokens": {
                "display_name": "Max tokens",
                "field_type": "int",
                "advanced": False,
                "required": False,
                "default": 256,
                "info": "The maximum number of tokens to generate for each chat completion.",
            },
            "max_retries": {
                "display_name": "Max retries",
                "field_type": "int",
                "advanced": True,
                "required": False,
                "default": 6,
            },
            "verbose": {
                "display_name": "Verbose",
                "field_type": "bool",
                "advanced": True,
                "required": False,
                "default": False,
            },
        }

    def build(
        self,
        model: str,
        api_key: str,
        streaming: bool = True,
        temperature: Optional[float] = 0.7,
        model_kwargs: Optional[Dict[str, Any]] = {},
        top_p: Optional[float] = None,
        top_k: Optional[int] = None,
        n: int = 1,
        max_tokens: int = 256,
        max_retries: int = 6,
        verbose: bool = False,
    ) -> Union[BaseLanguageModel, Callable]:
        try:
            import litellm  # type: ignore

            litellm.drop_params = True
            litellm.set_verbose = verbose
        except ImportError:
            raise ChatLiteLLMException(
                "Could not import litellm python package. " "Please install it with `pip install litellm`"
            )
        if api_key:
            if "perplexity" in model:
                os.environ["PERPLEXITYAI_API_KEY"] = api_key
            elif "replicate" in model:
                os.environ["REPLICATE_API_KEY"] = api_key

        LLM = ChatLiteLLM(
            model=model,
            client=None,
            streaming=streaming,
            temperature=temperature,
            model_kwargs=model_kwargs if model_kwargs is not None else {},
            top_p=top_p,
            top_k=top_k,
            n=n,
            max_tokens=max_tokens,
            max_retries=max_retries,
        )
        return LLM



import ast
import inspect
import operator
import traceback
from typing import Any, Dict, List, Type, Union

from cachetools import TTLCache, cachedmethod, keys
from fastapi import HTTPException


from langflow.interface.custom.schema import CallableCodeDetails, ClassCodeDetails


class CodeSyntaxError(HTTPException):
    pass


def get_data_type():
    from langflow.field_typing import Data

    return Data


def imports_key(*args, **kwargs):
    imports = kwargs.pop("imports")
    key = keys.methodkey(*args, **kwargs)
    key += tuple(imports)
    return key


class CodeParser:
    """
    A parser for Python source code, extracting code details.
    """

    def __init__(self, code: Union[str, Type]) -> None:
        """
        Initializes the parser with the provided code.
        """
        self.cache: TTLCache = TTLCache(maxsize=1024, ttl=60)
        if isinstance(code, type):
            if not inspect.isclass(code):
                raise ValueError("The provided code must be a class.")
            # If the code is a class, get its source code
            code = inspect.getsource(code)
        self.code = code
        self.data: Dict[str, Any] = {
            "imports": [],
            "functions": [],
            "classes": [],
            "global_vars": [],
        }
        self.handlers = {
            ast.Import: self.parse_imports,
            ast.ImportFrom: self.parse_imports,
            ast.FunctionDef: self.parse_functions,
            ast.ClassDef: self.parse_classes,
            ast.Assign: self.parse_global_vars,
        }

    def get_tree(self):
        """
        Parses the provided code to validate its syntax.
        It tries to parse the code into an abstract syntax tree (AST).
        """
        try:
            tree = ast.parse(self.code)
        except SyntaxError as err:
            raise CodeSyntaxError(
                status_code=400,
                detail={"error": err.msg, "traceback": traceback.format_exc()},
            ) from err

        return tree

    def parse_node(self, node: Union[ast.stmt, ast.AST]) -> None:
        """
        Parses an AST node and updates the data
        dictionary with the relevant information.
        """
        if handler := self.handlers.get(type(node)):  # type: ignore
            handler(node)  # type: ignore

    def parse_imports(self, node: Union[ast.Import, ast.ImportFrom]) -> None:
        """
        Extracts "imports" from the code, including aliases.
        """
        if isinstance(node, ast.Import):
            for alias in node.names:
                if alias.asname:
                    self.data["imports"].append(f"{alias.name} as {alias.asname}")
                else:
                    self.data["imports"].append(alias.name)
        elif isinstance(node, ast.ImportFrom):
            for alias in node.names:
                if alias.asname:
                    self.data["imports"].append((node.module, f"{alias.name} as {alias.asname}"))
                else:
                    self.data["imports"].append((node.module, alias.name))

    def parse_functions(self, node: ast.FunctionDef) -> None:
        """
        Extracts "functions" from the code.
        """
        self.data["functions"].append(self.parse_callable_details(node))

    def parse_arg(self, arg, default):
        """
        Parses an argument and its default value.
        """
        arg_dict = {"name": arg.arg, "default": default}
        if arg.annotation:
            arg_dict["type"] = ast.unparse(arg.annotation)
        return arg_dict

    @cachedmethod(operator.attrgetter("cache"))
    def construct_eval_env(self, return_type_str: str, imports) -> dict:
        """
        Constructs an evaluation environment with the necessary imports for the return type,
        taking into account module aliases.
        """
        eval_env: dict = {}
        for import_entry in imports:
            if isinstance(import_entry, tuple):  # from module import name
                module, name = import_entry
                if name in return_type_str:
                    exec(f"import {module}", eval_env)
                    exec(f"from {module} import {name}", eval_env)
            else:  # import module
                module = import_entry
                alias = None
                if " as " in module:
                    module, alias = module.split(" as ")
                if module in return_type_str or (alias and alias in return_type_str):
                    exec(f"import {module} as {alias if alias else module}", eval_env)
        return eval_env

    @cachedmethod(cache=operator.attrgetter("cache"))
    def parse_callable_details(self, node: ast.FunctionDef) -> Dict[str, Any]:
        """
        Extracts details from a single function or method node.
        """
        return_type = None
        if node.returns:
            return_type_str = ast.unparse(node.returns)
            eval_env = self.construct_eval_env(return_type_str, tuple(self.data["imports"]))

            try:
                return_type = eval(return_type_str, eval_env)
            except NameError:
                # Handle cases where the type is not found in the constructed environment
                pass

        func = CallableCodeDetails(
            name=node.name,
            doc=ast.get_docstring(node),
            args=self.parse_function_args(node),
            body=self.parse_function_body(node),
            return_type=return_type or get_data_type(),
            has_return=self.parse_return_statement(node),
        )

        return func.model_dump()

    def parse_function_args(self, node: ast.FunctionDef) -> List[Dict[str, Any]]:
        """
        Parses the arguments of a function or method node.
        """
        args = []

        args += self.parse_positional_args(node)
        args += self.parse_varargs(node)
        args += self.parse_keyword_args(node)
        # Commented out because we don't want kwargs
        # showing up as fields in the frontend
        # args += self.parse_kwargs(node)

        return args

    def parse_positional_args(self, node: ast.FunctionDef) -> List[Dict[str, Any]]:
        """
        Parses the positional arguments of a function or method node.
        """
        num_args = len(node.args.args)
        num_defaults = len(node.args.defaults)
        num_missing_defaults = num_args - num_defaults
        missing_defaults = [None] * num_missing_defaults
        default_values = [ast.unparse(default).strip("'") if default else None for default in node.args.defaults]
        # Now check all default values to see if there
        # are any "None" values in the middle
        default_values = [None if value == "None" else value for value in default_values]

        defaults = missing_defaults + default_values

        args = [self.parse_arg(arg, default) for arg, default in zip(node.args.args, defaults)]
        return args

    def parse_varargs(self, node: ast.FunctionDef) -> List[Dict[str, Any]]:
        """
        Parses the *args argument of a function or method node.
        """
        args = []

        if node.args.vararg:
            args.append(self.parse_arg(node.args.vararg, None))

        return args

    def parse_keyword_args(self, node: ast.FunctionDef) -> List[Dict[str, Any]]:
        """
        Parses the keyword-only arguments of a function or method node.
        """
        kw_defaults = [None] * (len(node.args.kwonlyargs) - len(node.args.kw_defaults)) + [
            ast.unparse(default) if default else None for default in node.args.kw_defaults
        ]

        args = [self.parse_arg(arg, default) for arg, default in zip(node.args.kwonlyargs, kw_defaults)]
        return args

    def parse_kwargs(self, node: ast.FunctionDef) -> List[Dict[str, Any]]:
        """
        Parses the **kwargs argument of a function or method node.
        """
        args = []

        if node.args.kwarg:
            args.append(self.parse_arg(node.args.kwarg, None))

        return args

    def parse_function_body(self, node: ast.FunctionDef) -> List[str]:
        """
        Parses the body of a function or method node.
        """
        return [ast.unparse(line) for line in node.body]

    def parse_return_statement(self, node: ast.FunctionDef) -> bool:
        """
        Parses the return statement of a function or method node, including nested returns.
        """

        def has_return(node):
            if isinstance(node, ast.Return):
                return True
            elif isinstance(node, ast.If):
                return any(has_return(child) for child in node.body) or any(has_return(child) for child in node.orelse)
            elif isinstance(node, ast.Try):
                return (
                    any(has_return(child) for child in node.body)
                    or any(has_return(child) for child in node.handlers)
                    or any(has_return(child) for child in node.finalbody)
                )
            elif isinstance(node, (ast.For, ast.While)):
                return any(has_return(child) for child in node.body) or any(has_return(child) for child in node.orelse)
            elif isinstance(node, ast.With):
                return any(has_return(child) for child in node.body)
            else:
                return False

        return any(has_return(child) for child in node.body)

    def parse_assign(self, stmt):
        """
        Parses an Assign statement and returns a dictionary
        with the target's name and value.
        """
        for target in stmt.targets:
            if isinstance(target, ast.Name):
                return {"name": target.id, "value": ast.unparse(stmt.value)}

    def parse_ann_assign(self, stmt):
        """
        Parses an AnnAssign statement and returns a dictionary
        with the target's name, value, and annotation.
        """
        if isinstance(stmt.target, ast.Name):
            return {
                "name": stmt.target.id,
                "value": ast.unparse(stmt.value) if stmt.value else None,
                "annotation": ast.unparse(stmt.annotation),
            }

    def parse_function_def(self, stmt):
        """
        Parses a FunctionDef statement and returns the parsed
        method and a boolean indicating if it's an __init__ method.
        """
        method = self.parse_callable_details(stmt)
        return (method, True) if stmt.name == "__init__" else (method, False)

    def parse_classes(self, node: ast.ClassDef) -> None:
        """
        Extracts "classes" from the code, including inheritance and init methods.
        """

        class_details = ClassCodeDetails(
            name=node.name,
            doc=ast.get_docstring(node),
            bases=[ast.unparse(base) for base in node.bases],
            attributes=[],
            methods=[],
            init=None,
        )

        for stmt in node.body:
            if isinstance(stmt, ast.Assign):
                if attr := self.parse_assign(stmt):
                    class_details.attributes.append(attr)
            elif isinstance(stmt, ast.AnnAssign):
                if attr := self.parse_ann_assign(stmt):
                    class_details.attributes.append(attr)
            elif isinstance(stmt, (ast.FunctionDef, ast.AsyncFunctionDef)):
                method, is_init = self.parse_function_def(stmt)
                if is_init:
                    class_details.init = method
                else:
                    class_details.methods.append(method)

        self.data["classes"].append(class_details.model_dump())

    def parse_global_vars(self, node: ast.Assign) -> None:
        """
        Extracts global variables from the code.
        """
        global_var = {
            "targets": [t.id if hasattr(t, "id") else ast.dump(t) for t in node.targets],
            "value": ast.unparse(node.value),
        }
        self.data["global_vars"].append(global_var)

    def parse_code(self) -> Dict[str, Any]:
        """
        Runs all parsing operations and returns the resulting data.
        """
        tree = self.get_tree()

        for node in ast.walk(tree):
            self.parse_node(node)
        return self.data



from datetime import datetime
from typing import List
from uuid import UUID

import orjson
from fastapi import APIRouter, Depends, File, HTTPException, UploadFile
from fastapi.encoders import jsonable_encoder
from sqlmodel import Session, select

from langflow.api.utils import remove_api_keys, validate_is_component
from langflow.api.v1.schemas import FlowListCreate, FlowListRead
from langflow.services.auth.utils import get_current_active_user
from langflow.services.database.models.flow import Flow, FlowCreate, FlowRead, FlowUpdate
from langflow.services.database.models.user.model import User
from langflow.services.deps import get_session, get_settings_service

# build router
router = APIRouter(prefix="/flows", tags=["Flows"])


@router.post("/", response_model=FlowRead, status_code=201)
def create_flow(
    *,
    session: Session = Depends(get_session),
    flow: FlowCreate,
    current_user: User = Depends(get_current_active_user),
):
    """Create a new flow."""
    if flow.user_id is None:
        flow.user_id = current_user.id

    db_flow = Flow.model_validate(flow, from_attributes=True)
    db_flow.updated_at = datetime.utcnow()

    session.add(db_flow)
    session.commit()
    session.refresh(db_flow)
    return db_flow


@router.get("/", response_model=list[FlowRead], status_code=200)
def read_flows(
    *,
    current_user: User = Depends(get_current_active_user),
):
    """Read all flows."""
    try:
        flows = current_user.flows
        flows = validate_is_component(flows)
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e)) from e
    return [jsonable_encoder(flow) for flow in flows]


@router.get("/{flow_id}", response_model=FlowRead, status_code=200)
def read_flow(
    *,
    session: Session = Depends(get_session),
    flow_id: UUID,
    current_user: User = Depends(get_current_active_user),
):
    """Read a flow."""
    if user_flow := (session.exec(select(Flow).where(Flow.id == flow_id, Flow.user_id == current_user.id)).first()):
        return user_flow
    else:
        raise HTTPException(status_code=404, detail="Flow not found")


@router.patch("/{flow_id}", response_model=FlowRead, status_code=200)
def update_flow(
    *,
    session: Session = Depends(get_session),
    flow_id: UUID,
    flow: FlowUpdate,
    current_user: User = Depends(get_current_active_user),
    settings_service=Depends(get_settings_service),
):
    """Update a flow."""

    db_flow = read_flow(session=session, flow_id=flow_id, current_user=current_user)
    if not db_flow:
        raise HTTPException(status_code=404, detail="Flow not found")
    flow_data = flow.model_dump(exclude_unset=True)
    if settings_service.settings.REMOVE_API_KEYS:
        flow_data = remove_api_keys(flow_data)
    for key, value in flow_data.items():
        if value is not None:
            setattr(db_flow, key, value)
    db_flow.updated_at = datetime.utcnow()
    session.add(db_flow)
    session.commit()
    session.refresh(db_flow)
    return db_flow


@router.delete("/{flow_id}", status_code=200)
def delete_flow(
    *,
    session: Session = Depends(get_session),
    flow_id: UUID,
    current_user: User = Depends(get_current_active_user),
):
    """Delete a flow."""
    flow = read_flow(session=session, flow_id=flow_id, current_user=current_user)
    if not flow:
        raise HTTPException(status_code=404, detail="Flow not found")
    session.delete(flow)
    session.commit()
    return {"message": "Flow deleted successfully"}


# Define a new model to handle multiple flows


@router.post("/batch/", response_model=List[FlowRead], status_code=201)
def create_flows(
    *,
    session: Session = Depends(get_session),
    flow_list: FlowListCreate,
    current_user: User = Depends(get_current_active_user),
):
    """Create multiple new flows."""
    db_flows = []
    for flow in flow_list.flows:
        flow.user_id = current_user.id
        db_flow = Flow.model_validate(flow, from_attributes=True)
        session.add(db_flow)
        db_flows.append(db_flow)
    session.commit()
    for db_flow in db_flows:
        session.refresh(db_flow)
    return db_flows


@router.post("/upload/", response_model=List[FlowRead], status_code=201)
async def upload_file(
    *,
    session: Session = Depends(get_session),
    file: UploadFile = File(...),
    current_user: User = Depends(get_current_active_user),
):
    """Upload flows from a file."""
    contents = await file.read()
    data = orjson.loads(contents)
    if "flows" in data:
        flow_list = FlowListCreate(**data)
    else:
        flow_list = FlowListCreate(flows=[FlowCreate(**flow) for flow in data])
    # Now we set the user_id for all flows
    for flow in flow_list.flows:
        flow.user_id = current_user.id

    return create_flows(session=session, flow_list=flow_list, current_user=current_user)


@router.get("/download/", response_model=FlowListRead, status_code=200)
async def download_file(
    *,
    session: Session = Depends(get_session),
    current_user: User = Depends(get_current_active_user),
):
    """Download all flows as a file."""
    flows = read_flows(current_user=current_user)
    return FlowListRead(flows=flows)



from typing import Dict, Optional

from langchain_community.embeddings.huggingface import HuggingFaceInferenceAPIEmbeddings
from langflow import CustomComponent
from pydantic.v1.types import SecretStr


class HuggingFaceInferenceAPIEmbeddingsComponent(CustomComponent):
    display_name = "HuggingFaceInferenceAPIEmbeddings"
    description = "HuggingFace sentence_transformers embedding models, API version."
    documentation = "https://github.com/huggingface/text-embeddings-inference"

    def build_config(self):
        return {
            "api_key": {"display_name": "API Key", "password": True, "advanced": True},
            "api_url": {"display_name": "API URL", "advanced": True},
            "model_name": {"display_name": "Model Name"},
            "cache_folder": {"display_name": "Cache Folder", "advanced": True},
            "encode_kwargs": {"display_name": "Encode Kwargs", "advanced": True, "field_type": "dict"},
            "model_kwargs": {"display_name": "Model Kwargs", "field_type": "dict", "advanced": True},
            "multi_process": {"display_name": "Multi Process", "advanced": True},
        }

    def build(
        self,
        api_key: Optional[str] = "",
        api_url: str = "http://localhost:8080",
        model_name: str = "BAAI/bge-large-en-v1.5",
        cache_folder: Optional[str] = None,
        encode_kwargs: Optional[Dict] = {},
        model_kwargs: Optional[Dict] = {},
        multi_process: bool = False,
    ) -> HuggingFaceInferenceAPIEmbeddings:
        if api_key:
            secret_api_key = SecretStr(api_key)
        else:
            raise ValueError("API Key is required")
        return HuggingFaceInferenceAPIEmbeddings(
            api_key=secret_api_key,
            api_url=api_url,
            model_name=model_name,
        )



import uuid
from datetime import datetime
from typing import Optional

from sqlmodel import Field, SQLModel


class Component(SQLModel, table=True):
    id: uuid.UUID = Field(default_factory=uuid.uuid4, primary_key=True)
    frontend_node_id: uuid.UUID = Field(index=True)
    name: str = Field(index=True)
    description: Optional[str] = Field(default=None)
    python_code: Optional[str] = Field(default=None)
    return_type: Optional[str] = Field(default=None)
    is_disabled: bool = Field(default=False)
    is_read_only: bool = Field(default=False)
    create_at: datetime = Field(default_factory=datetime.utcnow)
    update_at: datetime = Field(default_factory=datetime.utcnow)


class ComponentModel(SQLModel):
    id: uuid.UUID = Field(default_factory=uuid.uuid4)
    frontend_node_id: uuid.UUID = Field(default=uuid.uuid4())
    name: str = Field(default="")
    description: Optional[str] = None
    python_code: Optional[str] = None
    return_type: Optional[str] = None
    is_disabled: bool = False
    is_read_only: bool = False



from langchain_community.llms.cohere import Cohere
from langchain_core.language_models.base import BaseLanguageModel
from langflow import CustomComponent


class CohereComponent(CustomComponent):
    display_name = "Cohere"
    description = "Cohere large language models."
    documentation = "https://python.langchain.com/docs/modules/model_io/models/llms/integrations/cohere"

    def build_config(self):
        return {
            "cohere_api_key": {"display_name": "Cohere API Key", "type": "password", "password": True},
            "max_tokens": {"display_name": "Max Tokens", "default": 256, "type": "int", "show": True},
            "temperature": {"display_name": "Temperature", "default": 0.75, "type": "float", "show": True},
        }

    def build(
        self,
        cohere_api_key: str,
        max_tokens: int = 256,
        temperature: float = 0.75,
    ) -> BaseLanguageModel:
        return Cohere(cohere_api_key=cohere_api_key, max_tokens=max_tokens, temperature=temperature)  # type: ignore



import os
from typing import Any, Callable, Dict, Type

import orjson
from langchain_community.vectorstores import (
    FAISS,
    Chroma,
    ElasticsearchStore,
    MongoDBAtlasVectorSearch,
    Pinecone,
    Qdrant,
    SupabaseVectorStore,
    Weaviate,
)
from langchain_core.documents import Document


def docs_in_params(params: dict) -> bool:
    """Check if params has documents OR texts and one of them is not an empty list,
    If any of them is not an empty list, return True, else return False"""
    return ("documents" in params and params["documents"]) or ("texts" in params and params["texts"])


def initialize_mongodb(class_object: Type[MongoDBAtlasVectorSearch], params: dict):
    """Initialize mongodb and return the class object"""

    MONGODB_ATLAS_CLUSTER_URI = params.pop("mongodb_atlas_cluster_uri")
    if not MONGODB_ATLAS_CLUSTER_URI:
        raise ValueError("Mongodb atlas cluster uri must be provided in the params")
    import certifi
    from pymongo import MongoClient

    client: MongoClient = MongoClient(MONGODB_ATLAS_CLUSTER_URI, tlsCAFile=certifi.where())
    db_name = params.pop("db_name", None)
    collection_name = params.pop("collection_name", None)
    if not db_name or not collection_name:
        raise ValueError("db_name and collection_name must be provided in the params")

    index_name = params.pop("index_name", None)
    if not index_name:
        raise ValueError("index_name must be provided in the params")

    collection = client[db_name][collection_name]
    if not docs_in_params(params):
        # __init__ requires collection, embedding and index_name
        init_args = {
            "collection": collection,
            "index_name": index_name,
            "embedding": params.get("embedding"),
        }

        return class_object(**init_args)

    if "texts" in params:
        params["documents"] = params.pop("texts")

    params["collection"] = collection
    params["index_name"] = index_name

    return class_object.from_documents(**params)


def initialize_supabase(class_object: Type[SupabaseVectorStore], params: dict):
    """Initialize supabase and return the class object"""
    from supabase.client import Client, create_client

    if "supabase_url" not in params or "supabase_service_key" not in params:
        raise ValueError("Supabase url and service key must be provided in the params")
    if "texts" in params:
        params["documents"] = params.pop("texts")

    client_kwargs = {
        "supabase_url": params.pop("supabase_url"),
        "supabase_key": params.pop("supabase_service_key"),
    }

    supabase: Client = create_client(**client_kwargs)
    if not docs_in_params(params):
        params.pop("documents", None)
        params.pop("texts", None)
        return class_object(client=supabase, **params)
    # If there are docs in the params, create a new index

    return class_object.from_documents(client=supabase, **params)


def initialize_weaviate(class_object: Type[Weaviate], params: dict):
    """Initialize weaviate and return the class object"""
    if not docs_in_params(params):
        import weaviate  # type: ignore

        client_kwargs_json = params.get("client_kwargs", "{}")
        client_kwargs = orjson.loads(client_kwargs_json)
        client_params = {
            "url": params.get("weaviate_url"),
        }
        client_params.update(client_kwargs)
        weaviate_client = weaviate.Client(**client_params)

        new_params = {
            "client": weaviate_client,
            "index_name": params.get("index_name"),
            "text_key": params.get("text_key"),
        }
        return class_object(**new_params)
    # If there are docs in the params, create a new index
    if "texts" in params:
        params["documents"] = params.pop("texts")

    return class_object.from_documents(**params)


def initialize_faiss(class_object: Type[FAISS], params: dict):
    """Initialize faiss and return the class object"""

    if not docs_in_params(params):
        return class_object.load_local

    save_local = params.get("save_local")
    faiss_index = class_object(**params)
    if save_local:
        faiss_index.save_local(folder_path=save_local)
    return faiss_index


def initialize_pinecone(class_object: Type[Pinecone], params: dict):
    """Initialize pinecone and return the class object"""

    import pinecone  # type: ignore

    pinecone_api_key = params.pop("pinecone_api_key")
    pinecone_env = params.pop("pinecone_env")

    if pinecone_api_key is None or pinecone_env is None:
        if os.getenv("PINECONE_API_KEY") is not None:
            pinecone_api_key = os.getenv("PINECONE_API_KEY")
        if os.getenv("PINECONE_ENV") is not None:
            pinecone_env = os.getenv("PINECONE_ENV")

    if pinecone_api_key is None or pinecone_env is None:
        raise ValueError("Pinecone API key and environment must be provided in the params")

    # initialize pinecone
    pinecone.init(
        api_key=pinecone_api_key,  # find at app.pinecone.io
        environment=pinecone_env,  # next to api key in console
    )

    # If there are no docs in the params, return an existing index
    # but first remove any texts or docs keys from the params
    if not docs_in_params(params):
        existing_index_params = {
            "embedding": params.pop("embedding"),
        }
        if "index_name" in params:
            existing_index_params["index_name"] = params.pop("index_name")
        if "namespace" in params:
            existing_index_params["namespace"] = params.pop("namespace")

        return class_object.from_existing_index(**existing_index_params)
    # If there are docs in the params, create a new index
    if "texts" in params:
        params["documents"] = params.pop("texts")
    return class_object.from_documents(**params)


def initialize_chroma(class_object: Type[Chroma], params: dict):
    """Initialize a ChromaDB object from the params"""
    if (  # type: ignore
        "chroma_server_host" in params or "chroma_server_http_port" in params
    ):
        import chromadb  # type: ignore

        settings_params = {
            key: params[key] for key, value_ in params.items() if key.startswith("chroma_server_") and value_
        }
        chroma_settings = chromadb.config.Settings(**settings_params)
        params["client_settings"] = chroma_settings
    else:
        # remove all chroma_server_ keys from params
        params = {key: value for key, value in params.items() if not key.startswith("chroma_server_")}

    persist = params.pop("persist", False)
    if not docs_in_params(params):
        params.pop("documents", None)
        params.pop("texts", None)
        params["embedding_function"] = params.pop("embedding")
        chromadb_instance = class_object(**params)
    else:
        if "texts" in params:
            params["documents"] = params.pop("texts")
        for doc in params["documents"]:
            if not isinstance(doc, Document):
                # remove any non-Document objects from the list
                params["documents"].remove(doc)
                continue
            if doc.metadata is None:
                doc.metadata = {}
            for key, value in doc.metadata.items():
                if value is None:
                    doc.metadata[key] = ""

        chromadb_instance = class_object.from_documents(**params)
    if persist:
        chromadb_instance.persist()
    return chromadb_instance


def initialize_qdrant(class_object: Type[Qdrant], params: dict):
    if not docs_in_params(params):
        if "location" not in params and "api_key" not in params:
            raise ValueError("Location and API key must be provided in the params")
        from qdrant_client import QdrantClient

        client_params = {
            "location": params.pop("location"),
            "api_key": params.pop("api_key"),
        }
        lc_params = {
            "collection_name": params.pop("collection_name"),
            "embeddings": params.pop("embedding"),
        }
        client = QdrantClient(**client_params)

        return class_object(client=client, **lc_params)

    return class_object.from_documents(**params)


def initialize_elasticsearch(class_object: Type[ElasticsearchStore], params: dict):
    """Initialize elastic and return the class object"""
    if "index_name" not in params:
        raise ValueError("Elasticsearch Index must be provided in the params")
    if "es_url" not in params:
        raise ValueError("Elasticsearch URL must be provided in the params")
    if not docs_in_params(params):
        existing_index_params = {
            "embedding": params.pop("embedding"),
        }
        if "index_name" in params:
            existing_index_params["index_name"] = params.pop("index_name")
        if "es_url" in params:
            existing_index_params["es_url"] = params.pop("es_url")

        return class_object.from_existing_index(**existing_index_params)
    # If there are docs in the params, create a new index
    if "texts" in params:
        params["documents"] = params.pop("texts")
    return class_object.from_documents(**params)


vecstore_initializer: Dict[str, Callable[[Type[Any], dict], Any]] = {
    "Pinecone": initialize_pinecone,
    "Chroma": initialize_chroma,
    "Qdrant": initialize_qdrant,
    "Weaviate": initialize_weaviate,
    "ElasticsearchStore": initialize_elasticsearch,
    "FAISS": initialize_faiss,
    "SupabaseVectorStore": initialize_supabase,
    "MongoDBAtlasVectorSearch": initialize_mongodb,
}



from langflow.interface.custom.directory_reader import DirectoryReader
from langflow.template.frontend_node.custom_components import CustomComponentFrontendNode
from loguru import logger


def merge_nested_dicts_with_renaming(dict1, dict2):
    for key, value in dict2.items():
        if key in dict1 and isinstance(value, dict) and isinstance(dict1.get(key), dict):
            for sub_key, sub_value in value.items():
                # if sub_key in dict1[key]:
                #     new_key = get_new_key(dict1[key], sub_key)
                #     dict1[key][new_key] = sub_value
                # else:
                dict1[key][sub_key] = sub_value
        else:
            dict1[key] = value
    return dict1


def build_invalid_menu(invalid_components):
    """Build the invalid menu."""
    if not invalid_components.get("menu"):
        return {}

    logger.debug("------------------- INVALID COMPONENTS -------------------")
    invalid_menu = {}
    for menu_item in invalid_components["menu"]:
        menu_name = menu_item["name"]
        invalid_menu[menu_name] = build_invalid_menu_items(menu_item)
    return invalid_menu


def build_valid_menu(valid_components):
    """Build the valid menu."""
    valid_menu = {}
    logger.debug("------------------- VALID COMPONENTS -------------------")
    for menu_item in valid_components["menu"]:
        menu_name = menu_item["name"]
        valid_menu[menu_name] = build_menu_items(menu_item)
    return valid_menu


def build_and_validate_all_files(reader: DirectoryReader, file_list):
    """Build and validate all files"""
    data = reader.build_component_menu_list(file_list)

    valid_components = reader.filter_loaded_components(data=data, with_errors=False)
    invalid_components = reader.filter_loaded_components(data=data, with_errors=True)

    return valid_components, invalid_components


def load_files_from_path(path: str):
    """Load all files from a given path"""
    reader = DirectoryReader(path, False)

    return reader.get_files()


def build_custom_component_list_from_path(path: str):
    """Build a list of custom components for the langchain from a given path"""
    file_list = load_files_from_path(path)
    reader = DirectoryReader(path, False)

    valid_components, invalid_components = build_and_validate_all_files(reader, file_list)

    valid_menu = build_valid_menu(valid_components)
    invalid_menu = build_invalid_menu(invalid_components)

    return merge_nested_dicts_with_renaming(valid_menu, invalid_menu)


def create_invalid_component_template(component, component_name):
    """Create a template for an invalid component."""
    component_code = component["code"]
    component_frontend_node = CustomComponentFrontendNode(
        description="ERROR - Check your Python Code",
        display_name=f"ERROR - {component_name}",
    )

    component_frontend_node.error = component.get("error", None)
    field = component_frontend_node.template.get_field("code")
    field.value = component_code
    component_frontend_node.template.update_field("code", field)
    return component_frontend_node.model_dump(by_alias=True, exclude_none=True)


def log_invalid_component_details(component):
    """Log details of an invalid component."""
    logger.debug(component)
    logger.debug(f"Component Path: {component.get('path', None)}")
    logger.debug(f"Component Error: {component.get('error', None)}")


def build_invalid_component(component):
    """Build a single invalid component."""
    component_name = component["name"]
    component_template = create_invalid_component_template(component, component_name)
    log_invalid_component_details(component)
    return component_name, component_template


def build_invalid_menu_items(menu_item):
    """Build invalid menu items for a given menu."""
    menu_items = {}
    for component in menu_item["components"]:
        try:
            component_name, component_template = build_invalid_component(component)
            menu_items[component_name] = component_template
            logger.debug(f"Added {component_name} to invalid menu.")
        except Exception as exc:
            logger.exception(f"Error while creating custom component [{component_name}]: {str(exc)}")
    return menu_items


def get_new_key(dictionary, original_key):
    counter = 1
    new_key = original_key + " (" + str(counter) + ")"
    while new_key in dictionary:
        counter += 1
        new_key = original_key + " (" + str(counter) + ")"
    return new_key


def determine_component_name(component):
    """Determine the name of the component."""
    component_output_types = component["output_types"]
    if len(component_output_types) == 1:
        return component_output_types[0]
    else:
        file_name = component.get("file").split(".")[0]
        return "".join(word.capitalize() for word in file_name.split("_")) if "_" in file_name else file_name


def build_menu_items(menu_item):
    """Build menu items for a given menu."""
    menu_items = {}
    for component_name, component_template, component in menu_item["components"]:
        try:
            menu_items[component_name] = component_template
            logger.debug(f"Added {component_name} to valid menu.")
        except Exception as exc:
            logger.error(f"Error loading Component: {component['output_types']}")
            logger.exception(f"Error while building custom component {component['output_types']}: {exc}")
    return menu_items



from typing import TYPE_CHECKING

from langflow.services.cache.service import BaseCacheService, InMemoryCache, RedisCache
from langflow.services.factory import ServiceFactory
from langflow.utils.logger import logger

if TYPE_CHECKING:
    from langflow.services.settings.service import SettingsService


class CacheServiceFactory(ServiceFactory):
    def __init__(self):
        super().__init__(BaseCacheService)

    def create(self, settings_service: "SettingsService"):
        # Here you would have logic to create and configure a CacheService
        # based on the settings_service

        if settings_service.settings.CACHE_TYPE == "redis":
            logger.debug("Creating Redis cache")
            redis_cache = RedisCache(
                host=settings_service.settings.REDIS_HOST,
                port=settings_service.settings.REDIS_PORT,
                db=settings_service.settings.REDIS_DB,
                url=settings_service.settings.REDIS_URL,
                expiration_time=settings_service.settings.REDIS_CACHE_EXPIRE,
            )
            if redis_cache.is_connected():
                logger.debug("Redis cache is connected")
                return redis_cache
            logger.warning("Redis cache is not connected, falling back to in-memory cache")
            return InMemoryCache()

        elif settings_service.settings.CACHE_TYPE == "memory":
            return InMemoryCache()



from langchain.tools.retriever import create_retriever_tool

from langflow import CustomComponent
from langflow.field_typing import BaseRetriever, Tool


class RetrieverToolComponent(CustomComponent):
    display_name = "RetrieverTool"
    description = "Tool for interacting with retriever"

    def build_config(self):
        return {
            "retriever": {
                "display_name": "Retriever",
                "info": "Retriever to interact with",
                "type": BaseRetriever,
            },
            "name": {"display_name": "Name", "info": "Name of the tool"},
            "description": {"display_name": "Description", "info": "Description of the tool"},
        }

    def build(
        self,
        retriever: BaseRetriever,
        name: str,
        description: str,
    ) -> Tool:
        return create_retriever_tool(
            retriever=retriever,
            name=name,
            description=description,
        )



import re
from typing import ClassVar, Dict, Optional

from langflow.template.field.base import TemplateField
from langflow.template.frontend_node.constants import FORCE_SHOW_FIELDS
from langflow.template.frontend_node.formatter.base import FieldFormatter
from langflow.utils.constants import ANTHROPIC_MODELS, CHAT_OPENAI_MODELS, OPENAI_MODELS


class OpenAIAPIKeyFormatter(FieldFormatter):
    def format(self, field: TemplateField, name: Optional[str] = None) -> None:
        if field.name and "api_key" in field.name and "OpenAI" in str(name):
            field.display_name = "OpenAI API Key"
            field.required = False
            if field.value is None:
                field.value = ""


class ModelSpecificFieldFormatter(FieldFormatter):
    MODEL_DICT: ClassVar[Dict] = {
        "OpenAI": OPENAI_MODELS,
        "ChatOpenAI": CHAT_OPENAI_MODELS,
        "Anthropic": ANTHROPIC_MODELS,
        "ChatAnthropic": ANTHROPIC_MODELS,
    }

    def format(self, field: TemplateField, name: Optional[str] = None) -> None:
        if field.name and name in self.MODEL_DICT and field.name == "model_name":
            field.options = self.MODEL_DICT[name]
            field.is_list = True


class KwargsFormatter(FieldFormatter):
    def format(self, field: TemplateField, name: Optional[str] = None) -> None:
        if field.name and "kwargs" in field.name.lower():
            field.advanced = True
            field.required = False
            field.show = False


class APIKeyFormatter(FieldFormatter):
    def format(self, field: TemplateField, name: Optional[str] = None) -> None:
        if field.name and "api" in field.name.lower() and "key" in field.name.lower():
            field.required = False
            field.advanced = False

            field.display_name = (field.name or "").replace("_", " ").title()
            field.display_name = field.display_name.replace("Api", "API")


class RemoveOptionalFormatter(FieldFormatter):
    def format(self, field: TemplateField, name: Optional[str] = None) -> None:
        _type = field.field_type
        field.field_type = re.sub(r"Optional\[(.*)\]", r"\1", _type)


class ListTypeFormatter(FieldFormatter):
    def format(self, field: TemplateField, name: Optional[str] = None) -> None:
        _type = field.field_type
        is_list = "List" in _type or "Sequence" in _type
        if is_list:
            _type = re.sub(r"(List|Sequence)\[(.*)\]", r"\2", _type)
            field.is_list = True
        field.field_type = _type


class DictTypeFormatter(FieldFormatter):
    def format(self, field: TemplateField, name: Optional[str] = None) -> None:
        _type = field.field_type
        _type = _type.replace("Mapping", "dict")
        field.field_type = _type


class UnionTypeFormatter(FieldFormatter):
    def format(self, field: TemplateField, name: Optional[str] = None) -> None:
        _type = field.field_type
        if "Union" in _type:
            _type = _type.replace("Union[", "")[:-1]
            _type = _type.split(",")[0]
            _type = _type.replace("]", "").replace("[", "")
        field.field_type = _type


class SpecialFieldFormatter(FieldFormatter):
    SPECIAL_FIELD_HANDLERS: ClassVar[Dict] = {
        "allowed_tools": lambda field: "Tool",
        "max_value_length": lambda field: "int",
    }

    def format(self, field: TemplateField, name: Optional[str] = None) -> None:
        handler = self.SPECIAL_FIELD_HANDLERS.get(field.name)
        field.field_type = handler(field) if handler else field.field_type


class ShowFieldFormatter(FieldFormatter):
    def format(self, field: TemplateField, name: Optional[str] = None) -> None:
        key = field.name or ""
        required = field.required
        field.show = (
            (required and key not in ["input_variables"])
            or key in FORCE_SHOW_FIELDS
            or "api" in key
            or ("key" in key and "input" not in key and "output" not in key)
        )


class PasswordFieldFormatter(FieldFormatter):
    def format(self, field: TemplateField, name: Optional[str] = None) -> None:
        key = field.name or ""
        show = field.show
        if any(text in key.lower() for text in {"password", "token", "api", "key"}) and show:
            field.password = True


class MultilineFieldFormatter(FieldFormatter):
    def format(self, field: TemplateField, name: Optional[str] = None) -> None:
        key = field.name or ""
        if key in {
            "suffix",
            "prefix",
            "template",
            "examples",
            "code",
            "headers",
            "description",
        }:
            field.multiline = True


class DefaultValueFormatter(FieldFormatter):
    def format(self, field: TemplateField, name: Optional[str] = None) -> None:
        value = field.model_dump(by_alias=True, exclude_none=True)
        if "default" in value:
            field.value = value["default"]


class HeadersDefaultValueFormatter(FieldFormatter):
    def format(self, field: TemplateField, name: Optional[str] = None) -> None:
        key = field.name
        if key == "headers":
            field.value = """{"Authorization": "Bearer <token>"}"""


class DictCodeFileFormatter(FieldFormatter):
    def format(self, field: TemplateField, name: Optional[str] = None) -> None:
        key = field.name
        value = field.model_dump(by_alias=True, exclude_none=True)
        _type = value["type"]
        if "dict" in _type.lower() and key == "dict_":
            field.field_type = "file"
            field.file_types = [".json", ".yaml", ".yml"]
        elif _type.startswith("Dict") or _type.startswith("Mapping") or _type.startswith("dict"):
            field.field_type = "dict"



from abc import ABC, abstractmethod
from typing import Any, Callable


class TaskBackend(ABC):
    @abstractmethod
    def launch_task(self, task_func: Callable[..., Any], *args: Any, **kwargs: Any):
        pass

    @abstractmethod
    def get_task(self, task_id: str) -> Any:
        pass



from typing import Dict, List, Optional, Type

from langflow.interface.base import LangChainTypeCreator
from langflow.services.deps import get_settings_service
from langflow.template.frontend_node.textsplitters import TextSplittersFrontendNode
from langflow.interface.custom_lists import textsplitter_type_to_cls_dict

from loguru import logger
from langflow.utils.util import build_template_from_class


class TextSplitterCreator(LangChainTypeCreator):
    type_name: str = "textsplitters"

    @property
    def frontend_node_class(self) -> Type[TextSplittersFrontendNode]:
        return TextSplittersFrontendNode

    @property
    def type_to_loader_dict(self) -> Dict:
        return textsplitter_type_to_cls_dict

    def get_signature(self, name: str) -> Optional[Dict]:
        """Get the signature of a text splitter."""
        try:
            return build_template_from_class(name, textsplitter_type_to_cls_dict)
        except ValueError as exc:
            raise ValueError(f"Text Splitter {name} not found") from exc
        except AttributeError as exc:
            logger.error(f"Text Splitter {name} not loaded: {exc}")
            return None

    def to_list(self) -> List[str]:
        settings_service = get_settings_service()
        return [
            textsplitter.__name__
            for textsplitter in self.type_to_loader_dict.values()
            if textsplitter.__name__ in settings_service.settings.TEXTSPLITTERS or settings_service.settings.DEV
        ]


textsplitter_creator = TextSplitterCreator()



from typing import Optional

from langflow.template.field.base import TemplateField
from langflow.template.frontend_node.base import FrontendNode
from langflow.template.template.base import Template

DEFAULT_CUSTOM_COMPONENT_CODE = """from langflow import CustomComponent
from typing import Optional, List, Dict, Union
from langflow.field_typing import (
    AgentExecutor,
    BaseChatMemory,
    BaseLanguageModel,
    BaseLLM,
    BaseLoader,
    BaseMemory,
    BaseOutputParser,
    BasePromptTemplate,
    BaseRetriever,
    Callable,
    Chain,
    ChatPromptTemplate,
    Data,
    Document,
    Embeddings,
    NestedDict,
    Object,
    PromptTemplate,
    TextSplitter,
    Tool,
    VectorStore,
)


class Component(CustomComponent):
    display_name: str = "Custom Component"
    description: str = "Create any custom component you want!"

    def build_config(self):
        return {"param": {"display_name": "Parameter"}}

    def build(self, param: Data) -> Data:
        return param

"""


class CustomComponentFrontendNode(FrontendNode):
    _format_template: bool = False
    name: str = "CustomComponent"
    display_name: Optional[str] = "CustomComponent"
    beta: bool = True
    template: Template = Template(
        type_name="CustomComponent",
        fields=[
            TemplateField(
                field_type="code",
                required=True,
                placeholder="",
                is_list=False,
                show=True,
                value=DEFAULT_CUSTOM_COMPONENT_CODE,
                name="code",
                advanced=False,
                dynamic=True,
            )
        ],
    )
    description: Optional[str] = None
    base_classes: list[str] = []



from langflow import CustomComponent
from langchain.llms import BaseLLM
from typing import Optional, Union, Callable, Dict
from langchain_community.llms.vertexai import VertexAI


class VertexAIComponent(CustomComponent):
    display_name = "VertexAI"
    description = "Google Vertex AI large language models"

    def build_config(self):
        return {
            "credentials": {
                "display_name": "Credentials",
                "field_type": "file",
                "file_types": [".json"],
                "required": False,
                "value": None,
            },
            "location": {
                "display_name": "Location",
                "type": "str",
                "advanced": True,
                "value": "us-central1",
                "required": False,
            },
            "max_output_tokens": {
                "display_name": "Max Output Tokens",
                "field_type": "int",
                "value": 128,
                "required": False,
                "advanced": True,
            },
            "max_retries": {
                "display_name": "Max Retries",
                "type": "int",
                "value": 6,
                "required": False,
                "advanced": True,
            },
            "metadata": {
                "display_name": "Metadata",
                "field_type": "dict",
                "required": False,
                "default": {},
            },
            "model_name": {
                "display_name": "Model Name",
                "type": "str",
                "value": "text-bison",
                "required": False,
            },
            "n": {
                "advanced": True,
                "display_name": "N",
                "field_type": "int",
                "value": 1,
                "required": False,
            },
            "project": {
                "display_name": "Project",
                "type": "str",
                "required": False,
                "default": None,
            },
            "request_parallelism": {
                "display_name": "Request Parallelism",
                "field_type": "int",
                "value": 5,
                "required": False,
                "advanced": True,
            },
            "streaming": {
                "display_name": "Streaming",
                "field_type": "bool",
                "value": False,
                "required": False,
                "advanced": True,
            },
            "temperature": {
                "display_name": "Temperature",
                "field_type": "float",
                "value": 0.0,
                "required": False,
                "advanced": True,
            },
            "top_k": {"display_name": "Top K", "type": "int", "default": 40, "required": False, "advanced": True},
            "top_p": {
                "display_name": "Top P",
                "field_type": "float",
                "value": 0.95,
                "required": False,
                "advanced": True,
            },
            "tuned_model_name": {
                "display_name": "Tuned Model Name",
                "type": "str",
                "required": False,
                "value": None,
                "advanced": True,
            },
            "verbose": {
                "display_name": "Verbose",
                "field_type": "bool",
                "value": False,
                "required": False,
            },
            "name": {"display_name": "Name", "field_type": "str"},
        }

    def build(
        self,
        credentials: Optional[str] = None,
        location: str = "us-central1",
        max_output_tokens: int = 128,
        max_retries: int = 6,
        metadata: Dict = {},
        model_name: str = "text-bison",
        n: int = 1,
        name: Optional[str] = None,
        project: Optional[str] = None,
        request_parallelism: int = 5,
        streaming: bool = False,
        temperature: float = 0.0,
        top_k: int = 40,
        top_p: float = 0.95,
        tuned_model_name: Optional[str] = None,
        verbose: bool = False,
    ) -> Union[BaseLLM, Callable]:
        return VertexAI(
            credentials=credentials,
            location=location,
            max_output_tokens=max_output_tokens,
            max_retries=max_retries,
            metadata=metadata,
            model_name=model_name,
            n=n,
            name=name,
            project=project,
            request_parallelism=request_parallelism,
            streaming=streaming,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
            tuned_model_name=tuned_model_name,
            verbose=verbose,
        )



from langflow.services.deps import get_settings_service
from langflow.utils.lazy_load import LazyLoadDictBase


class AllTypesDict(LazyLoadDictBase):
    def __init__(self):
        self._all_types_dict = None

    @property
    def ALL_TYPES_DICT(self):
        return self.all_types_dict

    def _build_dict(self):
        langchain_types_dict = self.get_type_dict()
        return {
            **langchain_types_dict,
            "Custom": ["Custom Tool", "Python Function"],
        }

    def get_type_dict(self):
        from langflow.interface.types import get_all_types_dict

        settings_service = get_settings_service()
        return get_all_types_dict(settings_service=settings_service)


lazy_load_dict = AllTypesDict()



import contextlib
import re
from typing import Dict


def extract_input_variables(nodes):
    """
    Extracts input variables from the template
    and adds them to the input_variables field.
    """
    for node in nodes:
        with contextlib.suppress(Exception):
            if "input_variables" in node["data"]["node"]["template"]:
                if node["data"]["node"]["template"]["_type"] == "prompt":
                    variables = re.findall(
                        r"\{(.*?)\}",
                        node["data"]["node"]["template"]["template"]["value"],
                    )
                elif node["data"]["node"]["template"]["_type"] == "few_shot":
                    variables = re.findall(
                        r"\{(.*?)\}",
                        node["data"]["node"]["template"]["prefix"]["value"]
                        + node["data"]["node"]["template"]["suffix"]["value"],
                    )
                else:
                    variables = []
                node["data"]["node"]["template"]["input_variables"]["value"] = variables
    return nodes


def get_root_vertex(graph):
    """
    Returns the root node of the template.
    """
    incoming_edges = {edge.source_id for edge in graph.edges}

    if not incoming_edges and len(graph.vertices) == 1:
        return graph.vertices[0]

    return next((node for node in graph.vertices if node.id not in incoming_edges), None)


def build_json(root, graph) -> Dict:
    if "node" not in root.data:
        # If the root node has no "node" key, then it has only one child,
        # which is the target of the single outgoing edge
        edge = root.edges[0]
        local_nodes = [edge.target]
    else:
        # Otherwise, find all children whose type matches the type
        # specified in the template
        node_type = root.node_type
        local_nodes = graph.get_nodes_with_target(root)

    if len(local_nodes) == 1:
        return build_json(local_nodes[0], graph)
    # Build a dictionary from the template
    template = root.data["node"]["template"]
    final_dict = template.copy()

    for key, value in final_dict.items():
        if key == "_type":
            continue

        node_type = value["type"]

        if "value" in value and value["value"] is not None:
            # If the value is specified, use it
            value = value["value"]
        elif "dict" in node_type:
            # If the value is a dictionary, create an empty dictionary
            value = {}
        else:
            # Otherwise, recursively build the child nodes
            children = []
            for local_node in local_nodes:
                node_children = graph.get_children_by_node_type(local_node, node_type)
                children.extend(node_children)

            if value["required"] and not children:
                raise ValueError(f"No child with type {node_type} found")
            values = [build_json(child, graph) for child in children]
            value = (
                list(values) if value["list"] else next(iter(values), None)  # type: ignore
            )
        final_dict[key] = value

    return final_dict



from langflow.graph.edge.base import Edge
from langflow.graph.graph.base import Graph
from langflow.graph.vertex.base import Vertex
from langflow.graph.vertex.types import (
    AgentVertex,
    ChainVertex,
    DocumentLoaderVertex,
    EmbeddingVertex,
    LLMVertex,
    MemoryVertex,
    PromptVertex,
    TextSplitterVertex,
    ToolVertex,
    ToolkitVertex,
    VectorStoreVertex,
    WrapperVertex,
    RetrieverVertex,
)

__all__ = [
    "Graph",
    "Vertex",
    "Edge",
    "AgentVertex",
    "ChainVertex",
    "DocumentLoaderVertex",
    "EmbeddingVertex",
    "LLMVertex",
    "MemoryVertex",
    "PromptVertex",
    "TextSplitterVertex",
    "ToolVertex",
    "ToolkitVertex",
    "VectorStoreVertex",
    "WrapperVertex",
    "RetrieverVertex",
]



from typing import List, Optional
from uuid import UUID

from pydantic import BaseModel, validator


class TagResponse(BaseModel):
    id: UUID
    name: Optional[str]


class UsersLikesResponse(BaseModel):
    likes_count: Optional[int]
    liked_by_user: Optional[bool]


class CreateComponentResponse(BaseModel):
    id: UUID


class TagsIdResponse(BaseModel):
    tags_id: Optional[TagResponse]


class ListComponentResponse(BaseModel):
    id: Optional[UUID] = None
    name: Optional[str] = None
    description: Optional[str] = None
    liked_by_count: Optional[int] = None
    liked_by_user: Optional[bool] = None
    is_component: Optional[bool] = None
    metadata: Optional[dict] = {}
    user_created: Optional[dict] = {}
    tags: Optional[List[TagResponse]] = None
    downloads_count: Optional[int] = None
    last_tested_version: Optional[str] = None
    private: Optional[bool] = None

    # tags comes as a TagsIdResponse but we want to return a list of TagResponse
    @validator("tags", pre=True)
    def tags_to_list(cls, v):
        # Check if all values are have id and name
        # if so, return v else transform to TagResponse
        if not v:
            return v
        if all(["id" in tag and "name" in tag for tag in v]):
            return v
        else:
            return [TagResponse(**tag.get("tags_id")) for tag in v if tag.get("tags_id")]


class ListComponentResponseModel(BaseModel):
    count: Optional[int] = 0
    authorized: bool
    results: Optional[List[ListComponentResponse]]


class DownloadComponentResponse(BaseModel):
    id: UUID
    name: Optional[str]
    description: Optional[str]
    data: Optional[dict]
    is_component: Optional[bool]
    metadata: Optional[dict] = {}


class StoreComponentCreate(BaseModel):
    name: str
    description: Optional[str]
    data: dict
    tags: Optional[List[str]]
    parent: Optional[UUID] = None
    is_component: Optional[bool]
    last_tested_version: Optional[str] = None
    private: Optional[bool] = True



from typing import Optional
from langflow import CustomComponent
from langchain.retrievers import MetalRetriever
from langchain.schema import BaseRetriever
from metal_sdk.metal import Metal  # type: ignore


class MetalRetrieverComponent(CustomComponent):
    display_name: str = "Metal Retriever"
    description: str = "Retriever that uses the Metal API."

    def build_config(self):
        return {
            "api_key": {"display_name": "API Key", "password": True},
            "client_id": {"display_name": "Client ID", "password": True},
            "index_id": {"display_name": "Index ID"},
            "params": {"display_name": "Parameters"},
            "code": {"show": False},
        }

    def build(self, api_key: str, client_id: str, index_id: str, params: Optional[dict] = None) -> BaseRetriever:
        try:
            metal = Metal(api_key=api_key, client_id=client_id, index_id=index_id)
        except Exception as e:
            raise ValueError("Could not connect to Metal API.") from e
        return MetalRetriever(client=metal, params=params or {})



from typing import Optional
from langflow import CustomComponent
from langchain.retrievers import AmazonKendraRetriever
from langchain.schema import BaseRetriever


class AmazonKendraRetrieverComponent(CustomComponent):
    display_name: str = "Amazon Kendra Retriever"
    description: str = "Retriever that uses the Amazon Kendra API."

    def build_config(self):
        return {
            "index_id": {"display_name": "Index ID"},
            "region_name": {"display_name": "Region Name"},
            "credentials_profile_name": {"display_name": "Credentials Profile Name"},
            "attribute_filter": {
                "display_name": "Attribute Filter",
                "field_type": "code",
            },
            "top_k": {"display_name": "Top K", "field_type": "int"},
            "user_context": {
                "display_name": "User Context",
                "field_type": "code",
            },
            "code": {"show": False},
        }

    def build(
        self,
        index_id: str,
        top_k: int = 3,
        region_name: Optional[str] = None,
        credentials_profile_name: Optional[str] = None,
        attribute_filter: Optional[dict] = None,
        user_context: Optional[dict] = None,
    ) -> BaseRetriever:
        try:
            output = AmazonKendraRetriever(
                index_id=index_id,
                top_k=top_k,
                region_name=region_name,
                credentials_profile_name=credentials_profile_name,
                attribute_filter=attribute_filter,
                user_context=user_context,
            )  # type: ignore
        except Exception as e:
            raise ValueError("Could not connect to AmazonKendra API.") from e
        return output



from typing import Any, Dict, List, Optional, Type


from langflow.interface.base import LangChainTypeCreator

# from langflow.interface.custom.custom import CustomComponent
from langflow.interface.custom.custom_component import CustomComponent
from langflow.template.frontend_node.custom_components import (
    CustomComponentFrontendNode,
)
from loguru import logger

# Assuming necessary imports for Field, Template, and FrontendNode classes


class CustomComponentCreator(LangChainTypeCreator):
    type_name: str = "custom_components"

    @property
    def frontend_node_class(self) -> Type[CustomComponentFrontendNode]:
        return CustomComponentFrontendNode

    @property
    def type_to_loader_dict(self) -> Dict:
        if self.type_dict is None:
            self.type_dict: dict[str, Any] = {
                "CustomComponent": CustomComponent,
            }
        return self.type_dict

    def get_signature(self, name: str) -> Optional[Dict]:
        from langflow.custom.customs import get_custom_nodes

        try:
            if name in get_custom_nodes(self.type_name).keys():
                return get_custom_nodes(self.type_name)[name]
        except ValueError as exc:
            raise ValueError(f"CustomComponent {name} not found: {exc}") from exc
        except AttributeError as exc:
            logger.error(f"CustomComponent {name} not loaded: {exc}")
            return None
        return None

    def to_list(self) -> List[str]:
        return list(self.type_to_loader_dict.keys())


custom_component_creator = CustomComponentCreator()



import ast
import operator
import warnings
from typing import Any, ClassVar, Optional

import emoji
from cachetools import TTLCache, cachedmethod
from fastapi import HTTPException

from langflow.interface.custom.code_parser import CodeParser
from langflow.utils import validate


class ComponentCodeNullError(HTTPException):
    pass


class ComponentFunctionEntrypointNameNullError(HTTPException):
    pass


class Component:
    ERROR_CODE_NULL: ClassVar[str] = "Python code must be provided."
    ERROR_FUNCTION_ENTRYPOINT_NAME_NULL: ClassVar[str] = "The name of the entrypoint function must be provided."

    code: Optional[str] = None
    _function_entrypoint_name: str = "build"
    field_config: dict = {}
    _user_id: Optional[str]

    def __init__(self, **data):
        self.cache = TTLCache(maxsize=1024, ttl=60)
        for key, value in data.items():
            if key == "user_id":
                setattr(self, "_user_id", value)
            else:
                setattr(self, key, value)

        # Validate the emoji at the icon field
        if hasattr(self, "icon") and self.icon:
            self.icon = self.validate_icon(self.icon)

    def __setattr__(self, key, value):
        if key == "_user_id" and hasattr(self, "_user_id"):
            warnings.warn("user_id is immutable and cannot be changed.")
        super().__setattr__(key, value)

    @cachedmethod(cache=operator.attrgetter("cache"))
    def get_code_tree(self, code: str):
        parser = CodeParser(code)
        return parser.parse_code()

    def get_function(self):
        if not self.code:
            raise ComponentCodeNullError(
                status_code=400,
                detail={"error": self.ERROR_CODE_NULL, "traceback": ""},
            )

        if not self._function_entrypoint_name:
            raise ComponentFunctionEntrypointNameNullError(
                status_code=400,
                detail={
                    "error": self.ERROR_FUNCTION_ENTRYPOINT_NAME_NULL,
                    "traceback": "",
                },
            )

        return validate.create_function(self.code, self._function_entrypoint_name)

    def build_template_config(self, attributes) -> dict:
        template_config = {}

        for item in attributes:
            item_name = item.get("name")

            if item_value := item.get("value"):
                if "display_name" in item_name:
                    template_config["display_name"] = ast.literal_eval(item_value)

                elif "description" in item_name:
                    template_config["description"] = ast.literal_eval(item_value)

                elif "beta" in item_name:
                    template_config["beta"] = ast.literal_eval(item_value)

                elif "documentation" in item_name:
                    template_config["documentation"] = ast.literal_eval(item_value)

                elif "icon" in item_name:
                    icon_str = ast.literal_eval(item_value)
                    template_config["icon"] = self.validate_icon(icon_str)

        return template_config

    def validate_icon(self, value: str):
        # we are going to use the emoji library to validate the emoji
        # emojis can be defined using the :emoji_name: syntax
        if not value.startswith(":") or not value.endswith(":"):
            warnings.warn("Invalid emoji. Please use the :emoji_name: syntax.")
            return value
        emoji_value = emoji.emojize(value, variant="emoji_type")
        if value == emoji_value:
            warnings.warn(f"Invalid emoji. {value} is not a valid emoji.")
            return value
        return emoji_value

    def build(self, *args: Any, **kwargs: Any) -> Any:
        raise NotImplementedError



from typing import Callable, Optional, Union

from langchain.chains import LLMChain

from langflow import CustomComponent
from langflow.field_typing import (
    BaseLanguageModel,
    BaseMemory,
    BasePromptTemplate,
    Chain,
)


class LLMChainComponent(CustomComponent):
    display_name = "LLMChain"
    description = "Chain to run queries against LLMs"

    def build_config(self):
        return {
            "prompt": {"display_name": "Prompt"},
            "llm": {"display_name": "LLM"},
            "memory": {"display_name": "Memory"},
            "code": {"show": False},
        }

    def build(
        self,
        prompt: BasePromptTemplate,
        llm: BaseLanguageModel,
        memory: Optional[BaseMemory] = None,
    ) -> Union[Chain, Callable, LLMChain]:
        return LLMChain(prompt=prompt, llm=llm, memory=memory)



from contextlib import asynccontextmanager
from pathlib import Path
from typing import Optional
from urllib.parse import urlencode

from fastapi import FastAPI, Request
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import FileResponse
from fastapi.staticfiles import StaticFiles

from langflow.api import router
from langflow.interface.utils import setup_llm_caching
from langflow.services.plugins.langfuse_plugin import LangfuseInstance
from langflow.services.utils import initialize_services, teardown_services
from langflow.utils.logger import configure


@asynccontextmanager
async def lifespan(app: FastAPI):
    initialize_services()
    setup_llm_caching()
    LangfuseInstance.update()
    yield
    teardown_services()


def create_app():
    """Create the FastAPI app and include the router."""

    configure()

    app = FastAPI(lifespan=lifespan)
    origins = ["*"]

    app.add_middleware(
        CORSMiddleware,
        allow_origins=origins,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    @app.middleware("http")
    async def flatten_query_string_lists(request: Request, call_next):
        flattened: list[tuple[str, str]] = []
        for key, value in request.query_params.multi_items():
            flattened.extend((key, entry) for entry in value.split(","))

        request.scope["query_string"] = urlencode(flattened, doseq=True).encode("utf-8")

        return await call_next(request)

    @app.get("/health")
    def health():
        return {"status": "ok"}

    app.include_router(router)

    return app


def setup_static_files(app: FastAPI, static_files_dir: Path):
    """
    Setup the static files directory.
    Args:
        app (FastAPI): FastAPI app.
        path (str): Path to the static files directory.
    """
    app.mount(
        "/",
        StaticFiles(directory=static_files_dir, html=True),
        name="static",
    )

    @app.exception_handler(404)
    async def custom_404_handler(request, __):
        path = static_files_dir / "index.html"

        if not path.exists():
            raise RuntimeError(f"File at path {path} does not exist.")
        return FileResponse(path)


def get_static_files_dir():
    """Get the static files directory relative to Langflow's main.py file."""
    frontend_path = Path(__file__).parent
    return frontend_path / "frontend"


def setup_app(static_files_dir: Optional[Path] = None, backend_only: bool = False) -> FastAPI:
    """Setup the FastAPI app."""
    # get the directory of the current file
    if not static_files_dir:
        static_files_dir = get_static_files_dir()

    if not backend_only and (not static_files_dir or not static_files_dir.exists()):
        raise RuntimeError(f"Static files directory {static_files_dir} does not exist.")
    app = create_app()
    if not backend_only and static_files_dir is not None:
        setup_static_files(app, static_files_dir)
    return app


if __name__ == "__main__":
    import uvicorn

    from langflow.__main__ import get_number_of_workers

    configure()
    uvicorn.run(
        "langflow.main:create_app",
        host="127.0.0.1",
        port=7860,
        workers=get_number_of_workers(),
        log_level="debug",
        reload=True,
    )



from typing import Any

from langchain.agents import AgentExecutor
from langchain.chains.base import Chain
from langchain_core.runnables import Runnable
from langflow.api.v1.schemas import ChatMessage
from langflow.interface.utils import try_setting_streaming_options
from langflow.processing.base import get_result_and_steps
from loguru import logger

LANGCHAIN_RUNNABLES = (Chain, Runnable, AgentExecutor)


async def process_graph(
    build_result,
    chat_inputs: ChatMessage,
    client_id: str,
    session_id: str,
):
    build_result = try_setting_streaming_options(build_result)
    logger.debug("Loaded langchain object")

    if build_result is None:
        # Raise user facing error
        raise ValueError("There was an error loading the langchain_object. Please, check all the nodes and try again.")

    # Generate result and thought
    try:
        if chat_inputs.message is None:
            logger.debug("No message provided")
            chat_inputs.message = {}

        logger.debug("Generating result and thought")
        if isinstance(build_result, LANGCHAIN_RUNNABLES):
            result, intermediate_steps, raw_output = await get_result_and_steps(
                build_result,
                chat_inputs.message,
                client_id=client_id,
                session_id=session_id,
            )
        else:
            raise TypeError(f"Unknown type {type(build_result)}")
        logger.debug("Generated result and intermediate_steps")
        return result, intermediate_steps, raw_output
    except Exception as e:
        # Log stack trace
        logger.exception(e)
        raise e


async def run_build_result(build_result: Any, chat_inputs: ChatMessage, client_id: str, session_id: str):
    return build_result(inputs=chat_inputs.message)



from langflow import CustomComponent
from langflow.field_typing import BaseLanguageModel, AgentExecutor
from langchain_experimental.agents.agent_toolkits.csv.base import create_csv_agent


class CSVAgentComponent(CustomComponent):
    display_name = "CSVAgent"
    description = "Construct a CSV agent from a CSV and tools."
    documentation = "https://python.langchain.com/docs/modules/agents/toolkits/csv"

    def build_config(self):
        return {
            "llm": {"display_name": "LLM", "type": BaseLanguageModel},
            "path": {"display_name": "Path", "field_type": "file", "suffixes": [".csv"], "file_types": [".csv"]},
        }

    def build(
        self,
        llm: BaseLanguageModel,
        path: str,
    ) -> AgentExecutor:
        # Instantiate and return the CSV agent class with the provided llm and path
        return create_csv_agent(llm=llm, path=path)



# Router for base api
from fastapi import APIRouter

from langflow.api.v1 import (
    api_key_router,
    chat_router,
    credentials_router,
    endpoints_router,
    flows_router,
    login_router,
    store_router,
    users_router,
    validate_router,
)

router = APIRouter(
    prefix="/api/v1",
)
router.include_router(chat_router)
router.include_router(endpoints_router)
router.include_router(validate_router)
router.include_router(store_router)
router.include_router(flows_router)
router.include_router(users_router)
router.include_router(api_key_router)
router.include_router(login_router)
router.include_router(credentials_router)



from typing import Optional

from langchain.agents.mrkl import prompt

from langflow.template.field.base import TemplateField
from langflow.template.frontend_node.base import FrontendNode
from langflow.template.frontend_node.constants import DEFAULT_PROMPT, HUMAN_PROMPT, SYSTEM_PROMPT
from langflow.template.template.base import Template


class PromptFrontendNode(FrontendNode):
    @staticmethod
    def format_field(field: TemplateField, name: Optional[str] = None) -> None:
        FrontendNode.format_field(field, name)
        # if field.field_type  == "StringPromptTemplate"
        # change it to str
        PROMPT_FIELDS = [
            "template",
            "suffix",
            "prefix",
            "examples",
            "format_instructions",
        ]
        key = field.name or ""
        if field.field_type == "StringPromptTemplate" and "Message" in str(name):
            field.field_type = "prompt"
            field.multiline = True
            field.value = HUMAN_PROMPT if "Human" in key else SYSTEM_PROMPT
        if key == "template" and field.value == "":
            field.value = DEFAULT_PROMPT

        if key and key in PROMPT_FIELDS:
            field.field_type = "prompt"
            field.advanced = False

        if "Union" in field.field_type and "BaseMessagePromptTemplate" in field.field_type:
            field.field_type = "BaseMessagePromptTemplate"

        # All prompt fields should be password=False
        field.password = False
        field.dynamic = True


class PromptTemplateNode(FrontendNode):
    name: str = "PromptTemplate"
    template: Template
    description: str
    base_classes: list[str] = ["BasePromptTemplate"]

    @staticmethod
    def format_field(field: TemplateField, name: Optional[str] = None) -> None:
        FrontendNode.format_field(field, name)

        if (field.name or "") == "examples":
            field.advanced = False


class BasePromptFrontendNode(FrontendNode):
    name: str
    template: Template
    description: str
    base_classes: list[str]


class ZeroShotPromptNode(BasePromptFrontendNode):
    name: str = "ZeroShotPrompt"
    template: Template = Template(
        type_name="ZeroShotPrompt",
        fields=[
            TemplateField(
                field_type="str",
                required=False,
                placeholder="",
                is_list=False,
                show=True,
                multiline=True,
                value=prompt.PREFIX,
                name="prefix",
            ),
            TemplateField(
                field_type="str",
                required=True,
                placeholder="",
                is_list=False,
                show=True,
                multiline=True,
                value=prompt.FORMAT_INSTRUCTIONS,
                name="format_instructions",
            ),
            TemplateField(
                field_type="str",
                required=True,
                placeholder="",
                is_list=False,
                show=True,
                multiline=True,
                value=prompt.SUFFIX,
                name="suffix",
            ),
        ],
    )
    description: str = "Prompt template for Zero Shot Agent."
    base_classes: list[str] = ["BasePromptTemplate"]

    @staticmethod
    def format_field(field: TemplateField, name: Optional[str] = None) -> None:
        PromptFrontendNode.format_field(field, name)



from fastapi import APIRouter, Depends, HTTPException, Request, Response, status
from fastapi.security import OAuth2PasswordRequestForm
from sqlmodel import Session

from langflow.api.v1.schemas import Token
from langflow.services.auth.utils import (
    authenticate_user,
    create_refresh_token,
    create_user_longterm_token,
    create_user_tokens,
)
from langflow.services.deps import get_session, get_settings_service

router = APIRouter(tags=["Login"])


@router.post("/login", response_model=Token)
async def login_to_get_access_token(
    response: Response,
    form_data: OAuth2PasswordRequestForm = Depends(),
    db: Session = Depends(get_session),
    # _: Session = Depends(get_current_active_user)
    settings_service=Depends(get_settings_service),
):
    auth_settings = settings_service.auth_settings
    try:
        user = authenticate_user(form_data.username, form_data.password, db)
    except Exception as exc:
        if isinstance(exc, HTTPException):
            raise exc
        raise HTTPException(
            status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
            detail=str(exc),
        ) from exc

    if user:
        tokens = create_user_tokens(user_id=user.id, db=db, update_last_login=True)
        response.set_cookie(
            "refresh_token_lf",
            tokens["refresh_token"],
            httponly=auth_settings.REFRESH_HTTPONLY,
            samesite=auth_settings.REFRESH_SAME_SITE,
            secure=auth_settings.REFRESH_SECURE,
        )
        response.set_cookie(
            "access_token_lf",
            tokens["access_token"],
            httponly=auth_settings.ACCESS_HTTPONLY,
            samesite=auth_settings.ACCESS_SAME_SITE,
            secure=auth_settings.ACCESS_SECURE,
        )
        return tokens
    else:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Incorrect username or password",
            headers={"WWW-Authenticate": "Bearer"},
        )


@router.get("/auto_login")
async def auto_login(
    response: Response,
    db: Session = Depends(get_session),
    settings_service=Depends(get_settings_service),
):
    auth_settings = settings_service.auth_settings
    if settings_service.auth_settings.AUTO_LOGIN:
        tokens = create_user_longterm_token(db)
        response.set_cookie(
            "access_token_lf",
            tokens["access_token"],
            httponly=auth_settings.ACCESS_HTTPONLY,
            samesite=auth_settings.ACCESS_SAME_SITE,
            secure=auth_settings.ACCESS_SECURE,
        )
        return tokens

    raise HTTPException(
        status_code=status.HTTP_400_BAD_REQUEST,
        detail={
            "message": "Auto login is disabled. Please enable it in the settings",
            "auto_login": False,
        },
    )


@router.post("/refresh")
async def refresh_token(
    request: Request, response: Response, settings_service=Depends(get_settings_service)
):
    auth_settings = settings_service.auth_settings

    token = request.cookies.get("refresh_token_lf")

    if token:
        tokens = create_refresh_token(token)
        response.set_cookie(
            "refresh_token_lf",
            tokens["refresh_token"],
            httponly=auth_settings.REFRESH_HTTPONLY,
            samesite=auth_settings.REFRESH_SAME_SITE,
            secure=auth_settings.REFRESH_SECURE,
        )
        response.set_cookie(
            "access_token_lf",
            tokens["access_token"],
            httponly=auth_settings.ACCESS_HTTPONLY,
            samesite=auth_settings.ACCESS_SAME_SITE,
            secure=auth_settings.ACCESS_SECURE,
        )
        return tokens
    else:
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Invalid refresh token",
            headers={"WWW-Authenticate": "Bearer"},
        )


@router.post("/logout")
async def logout(response: Response):
    response.delete_cookie("refresh_token_lf")
    response.delete_cookie("access_token_lf")
    return {"message": "Logout successful"}



from typing import TYPE_CHECKING
from uuid import UUID

from fastapi import APIRouter, Depends, HTTPException
from sqlmodel import Session

from langflow.api.v1.schemas import ApiKeyCreateRequest, ApiKeysResponse
from langflow.services.auth import utils as auth_utils

# Assuming you have these methods in your service layer
from langflow.services.database.models.api_key.crud import (
    create_api_key,
    delete_api_key,
    get_api_keys,
)
from langflow.services.database.models.api_key.model import (
    ApiKeyCreate,
    UnmaskedApiKeyRead,
)
from langflow.services.database.models.user.model import User
from langflow.services.deps import (
    get_session,
    get_settings_service,
)

if TYPE_CHECKING:
    pass

router = APIRouter(tags=["APIKey"], prefix="/api_key")


@router.get("/", response_model=ApiKeysResponse)
def get_api_keys_route(
    db: Session = Depends(get_session),
    current_user: User = Depends(auth_utils.get_current_active_user),
):
    try:
        user_id = current_user.id
        keys = get_api_keys(db, user_id)

        return ApiKeysResponse(total_count=len(keys), user_id=user_id, api_keys=keys)
    except Exception as exc:
        raise HTTPException(status_code=400, detail=str(exc)) from exc


@router.post("/", response_model=UnmaskedApiKeyRead)
def create_api_key_route(
    req: ApiKeyCreate,
    current_user: User = Depends(auth_utils.get_current_active_user),
    db: Session = Depends(get_session),
):
    try:
        user_id = current_user.id
        return create_api_key(db, req, user_id=user_id)
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e)) from e


@router.delete("/{api_key_id}")
def delete_api_key_route(
    api_key_id: UUID,
    current_user=Depends(auth_utils.get_current_active_user),
    db: Session = Depends(get_session),
):
    try:
        delete_api_key(db, api_key_id)
        return {"detail": "API Key deleted"}
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e)) from e


@router.post("/store")
def save_store_api_key(
    api_key_request: ApiKeyCreateRequest,
    current_user: User = Depends(auth_utils.get_current_active_user),
    db: Session = Depends(get_session),
    settings_service=Depends(get_settings_service),
):
    try:
        api_key = api_key_request.api_key
        # Encrypt the API key
        encrypted = auth_utils.encrypt_api_key(api_key, settings_service=settings_service)
        current_user.store_api_key = encrypted
        db.commit()
        return {"detail": "API Key saved"}
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e)) from e


@router.delete("/store")
def delete_store_api_key(
    current_user: User = Depends(auth_utils.get_current_active_user),
    db: Session = Depends(get_session),
):
    try:
        current_user.store_api_key = None
        db.commit()
        return {"detail": "API Key deleted"}
    except Exception as e:
        raise HTTPException(status_code=400, detail=str(e)) from e



from typing import List, Optional

import requests
from langchain_core.documents import Document
from langflow import CustomComponent
from langflow.services.database.models.base import orjson_dumps


class UpdateRequest(CustomComponent):
    display_name: str = "Update Request"
    description: str = "Make a PATCH request to the given URL."
    output_types: list[str] = ["Document"]
    documentation: str = "https://docs.langflow.org/components/utilities#update-request"
    beta: bool = True
    field_config = {
        "url": {"display_name": "URL", "info": "The URL to make the request to."},
        "headers": {
            "display_name": "Headers",
            "field_type": "NestedDict",
            "info": "The headers to send with the request.",
        },
        "code": {"show": False},
        "document": {"display_name": "Document"},
        "method": {
            "display_name": "Method",
            "field_type": "str",
            "info": "The HTTP method to use.",
            "options": ["PATCH", "PUT"],
            "value": "PATCH",
        },
    }

    def update_document(
        self,
        session: requests.Session,
        document: Document,
        url: str,
        headers: Optional[dict] = None,
        method: str = "PATCH",
    ) -> Document:
        try:
            if method == "PATCH":
                response = session.patch(url, headers=headers, data=document.page_content)
            elif method == "PUT":
                response = session.put(url, headers=headers, data=document.page_content)
            else:
                raise ValueError(f"Unsupported method: {method}")
            try:
                response_json = response.json()
                result = orjson_dumps(response_json, indent_2=False)
            except Exception:
                result = response.text
            self.repr_value = result
            return Document(
                page_content=result,
                metadata={
                    "source": url,
                    "headers": headers,
                    "status_code": response.status_code,
                },
            )
        except Exception as exc:
            return Document(
                page_content=str(exc),
                metadata={"source": url, "headers": headers, "status_code": 500},
            )

    def build(
        self,
        method: str,
        document: Document,
        url: str,
        headers: Optional[dict] = None,
    ) -> List[Document]:
        if headers is None:
            headers = {}

        if not isinstance(document, list) and isinstance(document, Document):
            documents: list[Document] = [document]
        elif isinstance(document, list) and all(isinstance(doc, Document) for doc in document):
            documents = document
        else:
            raise ValueError("document must be a Document or a list of Documents")

        with requests.Session() as session:
            documents = [self.update_document(session, doc, url, headers, method) for doc in documents]
            self.repr_value = documents
        return documents



import traceback
from typing import Any, Callable, Optional, Tuple

import anyio
from loguru import logger

from langflow.services.task.backends.base import TaskBackend


class AnyIOTaskResult:
    def __init__(self, scope):
        self._scope = scope
        self._status = "PENDING"
        self._result = None
        self._exception = None

    @property
    def status(self) -> str:
        if self._status == "DONE":
            return "FAILURE" if self._exception is not None else "SUCCESS"
        return self._status

    @property
    def traceback(self) -> str:
        if self._traceback is not None:
            return "".join(traceback.format_tb(self._traceback))
        return ""

    @property
    def result(self) -> Any:
        return self._result

    def ready(self) -> bool:
        return self._status == "DONE"

    async def run(self, func, *args, **kwargs):
        try:
            self._result = await func(*args, **kwargs)
        except Exception as e:
            self._exception = e
            self._traceback = e.__traceback__
        finally:
            self._status = "DONE"


class AnyIOBackend(TaskBackend):
    name = "anyio"

    def __init__(self):
        self.tasks = {}

    async def launch_task(
        self, task_func: Callable[..., Any], *args: Any, **kwargs: Any
    ) -> Tuple[Optional[str], Optional[AnyIOTaskResult]]:
        """
        Launch a new task in an asynchronous manner.

        Parameters:
            task_func: The asynchronous function to run.
            *args: Positional arguments to pass to task_func.
            **kwargs: Keyword arguments to pass to task_func.

        Returns:
            A tuple containing a unique task ID and the task result object.
        """
        async with anyio.create_task_group() as tg:
            try:
                task_result = AnyIOTaskResult(tg)
                tg.start_soon(task_result.run, task_func, *args, **kwargs)
                task_id = str(id(task_result))
                self.tasks[task_id] = task_result
                logger.info(f"Task {task_id} started.")
                return task_id, task_result
            except Exception as e:
                logger.error(f"An error occurred while launching the task: {e}")
                return None, None

    def get_task(self, task_id: str) -> Any:
        return self.tasks.get(task_id)



from typing import List, Optional, Union

from langchain.llms import BaseLLM
from langchain_community.chat_models.vertexai import ChatVertexAI
from langchain_core.messages.base import BaseMessage
from langflow import CustomComponent
from langflow.field_typing import BaseLanguageModel


class ChatVertexAIComponent(CustomComponent):
    display_name = "ChatVertexAI"
    description = "`Vertex AI` Chat large language models API."

    def build_config(self):
        return {
            "credentials": {
                "display_name": "Credentials",
                "field_type": "file",
                "file_types": [".json"],
                "file_path": None,
            },
            "examples": {
                "display_name": "Examples",
                "multiline": True,
            },
            "location": {
                "display_name": "Location",
                "value": "us-central1",
            },
            "max_output_tokens": {
                "display_name": "Max Output Tokens",
                "value": 128,
                "advanced": True,
            },
            "model_name": {
                "display_name": "Model Name",
                "value": "chat-bison",
            },
            "project": {
                "display_name": "Project",
            },
            "temperature": {
                "display_name": "Temperature",
                "value": 0.0,
            },
            "top_k": {
                "display_name": "Top K",
                "value": 40,
                "advanced": True,
            },
            "top_p": {
                "display_name": "Top P",
                "value": 0.95,
                "advanced": True,
            },
            "verbose": {
                "display_name": "Verbose",
                "value": False,
                "advanced": True,
            },
        }

    def build(
        self,
        credentials: Optional[str],
        project: str,
        examples: Optional[List[BaseMessage]] = [],
        location: str = "us-central1",
        max_output_tokens: int = 128,
        model_name: str = "chat-bison",
        temperature: float = 0.0,
        top_k: int = 40,
        top_p: float = 0.95,
        verbose: bool = False,
    ) -> Union[BaseLanguageModel, BaseLLM]:
        return ChatVertexAI(
            credentials=credentials,
            examples=examples,
            location=location,
            max_output_tokens=max_output_tokens,
            model_name=model_name,
            project=project,
            temperature=temperature,
            top_k=top_k,
            top_p=top_p,
            verbose=verbose,
        )



from typing import Any, Optional

from pydantic import BaseModel, Field


class ClassCodeDetails(BaseModel):
    """
    A dataclass for storing details about a class.
    """

    name: str
    doc: Optional[str] = None
    bases: list
    attributes: list
    methods: list
    init: Optional[dict] = Field(default_factory=dict)


class CallableCodeDetails(BaseModel):
    """
    A dataclass for storing details about a callable.
    """

    name: str
    doc: Optional[str] = None
    args: list
    body: list
    return_type: Optional[Any] = None
    has_return: bool = False



import tempfile
import urllib
import urllib.request
from typing import List, Optional, Union

from langchain_community.embeddings import FakeEmbeddings
from langchain_community.vectorstores.vectara import Vectara
from langchain_core.vectorstores import VectorStore
from langflow import CustomComponent
from langflow.field_typing import BaseRetriever, Document


class VectaraComponent(CustomComponent):
    display_name: str = "Vectara"
    description: str = "Implementation of Vector Store using Vectara"
    documentation = "https://python.langchain.com/docs/integrations/vectorstores/vectara"
    beta = True
    field_config = {
        "vectara_customer_id": {
            "display_name": "Vectara Customer ID",
        },
        "vectara_corpus_id": {
            "display_name": "Vectara Corpus ID",
        },
        "vectara_api_key": {
            "display_name": "Vectara API Key",
            "password": True,
        },
        "documents": {"display_name": "Documents", "info": "If provided, will be upserted to corpus (optional)"},
        "files_url": {
            "display_name": "Files Url",
            "info": "Make vectara object using url of files (optional)",
        },
    }

    def build(
        self,
        vectara_customer_id: str,
        vectara_corpus_id: str,
        vectara_api_key: str,
        files_url: Optional[List[str]] = None,
        documents: Optional[Document] = None,
    ) -> Union[VectorStore, BaseRetriever]:
        source = "Langflow"

        if documents is not None:
            return Vectara.from_documents(
                documents=documents,  # type: ignore
                embedding=FakeEmbeddings(size=768),
                vectara_customer_id=vectara_customer_id,
                vectara_corpus_id=vectara_corpus_id,
                vectara_api_key=vectara_api_key,
                source=source,
            )

        if files_url is not None:
            files_list = []
            for url in files_url:
                name = tempfile.NamedTemporaryFile().name
                urllib.request.urlretrieve(url, name)
                files_list.append(name)

            return Vectara.from_files(
                files=files_list,
                embedding=FakeEmbeddings(size=768),
                vectara_customer_id=vectara_customer_id,
                vectara_corpus_id=vectara_corpus_id,
                vectara_api_key=vectara_api_key,
                source=source,
            )

        return Vectara(
            vectara_customer_id=vectara_customer_id,
            vectara_corpus_id=vectara_corpus_id,
            vectara_api_key=vectara_api_key,
            source=source,
        )



from typing import Any, Callable, Coroutine, Union

from langflow.services.base import Service
from langflow.services.task.backends.anyio import AnyIOBackend
from langflow.services.task.backends.base import TaskBackend
from langflow.services.task.utils import get_celery_worker_status
from langflow.utils.logger import configure
from loguru import logger


def check_celery_availability():
    try:
        from langflow.worker import celery_app

        status = get_celery_worker_status(celery_app)
        logger.debug(f"Celery status: {status}")
    except Exception as exc:
        logger.debug(f"Celery not available: {exc}")
        status = {"availability": None}
    return status


try:
    configure()
    status = check_celery_availability()

    USE_CELERY = status.get("availability") is not None
except ImportError:
    USE_CELERY = False


class TaskService(Service):
    name = "task_service"

    def __init__(self):
        self.backend = self.get_backend()
        self.use_celery = USE_CELERY

    @property
    def backend_name(self) -> str:
        return self.backend.name

    def get_backend(self) -> TaskBackend:
        if USE_CELERY:
            from langflow.services.task.backends.celery import CeleryBackend

            logger.debug("Using Celery backend")
            return CeleryBackend()
        logger.debug("Using AnyIO backend")
        return AnyIOBackend()

    # In your TaskService class
    async def launch_and_await_task(
        self,
        task_func: Callable[..., Any],
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        if not self.use_celery:
            return None, await task_func(*args, **kwargs)
        if not hasattr(task_func, "apply"):
            raise ValueError(f"Task function {task_func} does not have an apply method")
        task = task_func.apply(args=args, kwargs=kwargs)

        result = task.get()
        # if result is coroutine
        if isinstance(result, Coroutine):
            result = await result
        return task.id, result

    async def launch_task(self, task_func: Callable[..., Any], *args: Any, **kwargs: Any) -> Any:
        logger.debug(f"Launching task {task_func} with args {args} and kwargs {kwargs}")
        logger.debug(f"Using backend {self.backend}")
        task = self.backend.launch_task(task_func, *args, **kwargs)
        return await task if isinstance(task, Coroutine) else task

    def get_task(self, task_id: Union[int, str]) -> Any:
        return self.backend.get_task(task_id)



from langchain import tools
from langchain.agents import Tool
from langchain.agents.load_tools import _BASE_TOOLS, _EXTRA_LLM_TOOLS, _EXTRA_OPTIONAL_TOOLS, _LLM_TOOLS
from langchain.tools.json.tool import JsonSpec
from langflow.interface.importing.utils import import_class
from langflow.interface.tools.custom import PythonFunction, PythonFunctionTool

FILE_TOOLS = {"JsonSpec": JsonSpec}
CUSTOM_TOOLS = {
    "Tool": Tool,
    "PythonFunctionTool": PythonFunctionTool,
    "PythonFunction": PythonFunction,
}

OTHER_TOOLS = {tool: import_class(f"langchain_community.tools.{tool}") for tool in tools.__all__}

ALL_TOOLS_NAMES = {
    **_BASE_TOOLS,
    **_LLM_TOOLS,  # type: ignore
    **{k: v[0] for k, v in _EXTRA_LLM_TOOLS.items()},  # type: ignore
    **{k: v[0] for k, v in _EXTRA_OPTIONAL_TOOLS.items()},
    **CUSTOM_TOOLS,
    **FILE_TOOLS,  # type: ignore
    **OTHER_TOOLS,
}



from typing import Callable, Optional, Union

from langchain.chains.combine_documents.base import BaseCombineDocumentsChain
from langchain.chains.retrieval_qa.base import BaseRetrievalQA, RetrievalQA
from langflow import CustomComponent
from langflow.field_typing import BaseMemory, BaseRetriever


class RetrievalQAComponent(CustomComponent):
    display_name = "RetrievalQA"
    description = "Chain for question-answering against an index."

    def build_config(self):
        return {
            "combine_documents_chain": {"display_name": "Combine Documents Chain"},
            "retriever": {"display_name": "Retriever"},
            "memory": {"display_name": "Memory", "required": False},
            "input_key": {"display_name": "Input Key", "advanced": True},
            "output_key": {"display_name": "Output Key", "advanced": True},
            "return_source_documents": {"display_name": "Return Source Documents"},
        }

    def build(
        self,
        combine_documents_chain: BaseCombineDocumentsChain,
        retriever: BaseRetriever,
        memory: Optional[BaseMemory] = None,
        input_key: str = "query",
        output_key: str = "result",
        return_source_documents: bool = True,
    ) -> Union[BaseRetrievalQA, Callable]:
        return RetrievalQA(
            combine_documents_chain=combine_documents_chain,
            retriever=retriever,
            memory=memory,
            input_key=input_key,
            output_key=output_key,
            return_source_documents=return_source_documents,
        )



from enum import Enum


class CredentialType(str, Enum):
    """CredentialType is an Enum of the accepted providers"""

    OPENAI_API_KEY = "OPENAI_API_KEY"
    ANTHROPIC_API_KEY = "ANTHROPIC_API_KEY"



from langflow.template.field.base import TemplateField
from langflow.template.frontend_node.base import FrontendNode
from langchain.text_splitter import Language


class TextSplittersFrontendNode(FrontendNode):
    def add_extra_base_classes(self) -> None:
        self.base_classes = ["Document"]
        self.output_types = ["Document"]

    def add_extra_fields(self) -> None:
        self.template.add_field(
            TemplateField(
                field_type="Document",
                required=True,
                show=True,
                name="documents",
                is_list=True,
            )
        )
        name = "separator"
        if self.template.type_name == "CharacterTextSplitter":
            name = "separator"
        elif self.template.type_name == "RecursiveCharacterTextSplitter":
            name = "separators"
            # Add a field for type of separator
            # which will have Text or any value from the
            # Language enum
            options = [x.value for x in Language] + ["Text"]
            options.sort()
            self.template.add_field(
                TemplateField(
                    field_type="str",
                    required=True,
                    show=True,
                    name="separator_type",
                    advanced=False,
                    is_list=True,
                    options=options,
                    value="Text",
                    display_name="Separator Type",
                )
            )
        self.template.add_field(
            TemplateField(
                field_type="str",
                required=True,
                show=True,
                value="\\n",
                name=name,
                display_name="Separator",
            )
        )
        self.template.add_field(
            TemplateField(
                field_type="int",
                required=True,
                show=True,
                value=1000,
                name="chunk_size",
                display_name="Chunk Size",
            )
        )
        self.template.add_field(
            TemplateField(
                field_type="int",
                required=True,
                show=True,
                value=200,
                name="chunk_overlap",
                display_name="Chunk Overlap",
            )
        )
